#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                ampel/logging/AmpelLoggingError.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.12.2018
# Last Modified Date:  18.01.2019
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

class AmpelLoggingError(Exception):
	pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/T2Command.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.03.2021
# Last Modified Date:  27.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from json import dumps
from datetime import datetime
from argparse import ArgumentParser
from typing import Any
from collections.abc import Sequence
from ampel.core.EventHandler import EventHandler
from ampel.core.AmpelContext import AmpelContext
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.enum.DocumentCode import DocumentCode
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag
from ampel.t2.T2Utils import T2Utils
from ampel.util.pretty import prettyjson
from ampel.util.serialize import walk_and_encode
from ampel.content.T2Document import T2Document
from ampel.cli.utils import maybe_load_idmapper, maybe_resolve_enum
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.cli.LoadJSONAction import LoadJSONAction
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser

hlp = {
	"show": "Show T2 document(s) as JSON (stdout)",
	"save": "Save T2 document(s) as JSON using the provide file path",
	"reset": "Delete body and set the run state to NEW",
	"soft-reset": "Set the run state to NEW",
	"config": "Path to an ampel config file (yaml/json)",
	"secrets": "Path to a YAML secrets store in sops format",
	'out': 'Path to file were output will be written (printed to stdout otherwise)',
	'db': 'Database prefix. If set, "-mongo.prefix" value will be ignored',
	"unit": "Unit id/name",
	"limit": "Limit number of returned documents",
	"unit-config": "Unit config (integer number). Use string 'null' to match null",
	"code": "T2 code (0: COMPLETED, -1: NEW)",
	"link": "T2 document link (hex)",
	"stock": "Stock(s) associated with t2 doc (OR matched if multi-valued)",
	'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
	"custom-match": "Custom mongodb match as JSON string (ex: {\"body.result.mychi2\": {\"$gt\": 1.0}})",
	'no-resolve-stock': 'Keep stock as int when matching using id-mapper',
	"resolve-config": "Translate 'config' field from int back to dict",
	"human-times": "Translate timestamps to human-readable strings",
	"pretty-json": "Prettify JSON output"
}

class T2Command(AbsCoreCommand):

	@staticmethod
	def get_sub_ops() -> list[str]:
		return ['show', 'save', 'reset', 'soft-reset']

	# Mandatory implementation
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = self.get_sub_ops()
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				't2', sub_ops, hlp,
				description = 'Allows to match and then view or reset T2 documents.\n' +
				'Reset operations are registered in the event database and a\n' +
				'dedicated journal entry is added to the affected t2 documents.'
			)

		builder = ArgParserBuilder("t2")
		builder.add_parsers(sub_ops, hlp)

		builder.notation_add_note_references()
		builder.notation_add_example_references()

		# Required
		builder.opt('config')
		builder.req('out', sub_ops='save')

		# Optional general
		builder.opt('secrets', default=None)
		builder.opt('db', default=None)
		builder.opt('id-mapper', sub_ops='show|save')
		builder.opt('dry-run', action='store_true')
		builder.opt('debug', action='count', default=0, help='Debug')
		builder.opt('limit', default=None, type=int, sub_ops='show|save')
		builder.opt('pretty-json', sub_ops='show|save', action='store_true')

		# Optional match criteria
		builder.add_group('match', 'Optional T2 documents matching criteria', sub_ops='all')
		builder.arg('unit', group='match', nargs='+', action=MaybeIntAction)
		builder.arg('unit-config', group='match', nargs='+', action=MaybeIntAction)
		builder.arg('code', group='match', nargs='+', action=MaybeIntAction)
		builder.arg('link', group='match', nargs='+')
		builder.arg('stock', group='match', nargs='+', action=MaybeIntAction)
		builder.logic_args('run', descr='Run id', group='match', pos=0, ref='2')
		builder.logic_args('channel', descr='Channel', group='match', ref='2')
		builder.logic_args('with-tag', descr='Tag', group='match', ref='2', json=False)
		builder.logic_args('without-tag', descr='Tag', group='match', excl=True, ref='2', json=False)
		builder.arg('custom-match', group='match', metavar='#', action=LoadJSONAction)

		builder.add_group('format', 'Optional format parameters', sub_ops='show|save')
		builder.arg('resolve-config', group='format', sub_ops='show|save', action='store_true')
		builder.arg('human-times', group='format', sub_ops='show|save', action='store_true')
		builder.arg('no-resolve-stock', group='format', sub_ops='show|save', action='store_true')

		builder.note(
			'Reset operations failing to match any t2 document will not be registered in\n' +
			'in the event database but these will sliently increase the global run counter',
			sub_ops='reset|soft-reset'
		)

		# Examples
		for el in sub_ops:
			p = f"ampel t2 {el} -config ampel_conf.yaml "
			a = " -out /path/to/file" if el == "save" else ""
			builder.example(el, '-unit T2SNCosmo -db AmpelTest', prepend=p, append=a)
			builder.example(el, '-channel MY_CHANNEL -code -1', prepend=p, append=a)
			builder.example(el, '-stock 122621027 122620210 -unit DemoTiedLightCurveT2Unit', prepend=p, append=a)

		builder.example(
			'show|save',
			'-db AmpelTest -human-times -stock ZTF20aaqubac -resolve-config -id-mapper ZTFIdMapper'
		)

		self.parsers = builder.get()
		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		if sub_op is None:
			raise ValueError("A sub-operation (show, save, reset, soft-reset) needs to be specified")

		ctx = self.get_context(
			args, unknown_args, ContextClass=AmpelContext,
			require_existing_db = args['db'] or True, one_db='auto'
		)

		logger = AmpelLogger.from_profile(
			ctx, 'console_debug' if args['debug'] else 'console_info',
			base_flag=LogFlag.MANUAL_RUN
		)

		args['config'] = args.pop('unit_config')

		t2_utils = T2Utils(logger)
		col = ctx.db.get_collection('t2', mode='r')

		maybe_load_idmapper(args)
		if args['code'] is not None:
			args['code'] = maybe_resolve_enum(args['code'], DocumentCode)
		self.convert_logical_args('tag', args)

		# args['id_mapper'] is used for matching whereas id_mapper is potentially discarded for printing
		id_mapper = None if args.get('no_resolve_stock') else args['id_mapper']
		jsondump: Any = prettyjson if args.get('pretty_json') else lambda x: dumps(walk_and_encode(x)) # type: ignore

		if sub_op == 'show':

			m = t2_utils.match_t2s(**args)
			limit = args.get('limit', None)
			if args.get('dry_run'):
				count = col.count_documents(m)
				if limit:
					count = min(limit, count)
				logger.info(
					f"Query: {m}\n"
					f"Number of matched documents: {count}\n"
					f"Exiting (dry-run)"
				)
				return

			c = col.find(m).limit(limit) if limit is not None else col.find(m)

			if args['resolve_config'] or args['human_times'] or id_mapper:
				resolve_config = args['resolve_config']
				human_times = args['human_times']
				for el in c:
					self.morph_ret(ctx, el, resolve_config, human_times, id_mapper)
					print(jsondump(el))
			else:
				print(jsondump(list(c)))

		elif sub_op == 'save':

			cc = t2_utils.get_t2s(col=col, **args)
			if args.get('dry_run'):
				logger.info("Exiting (dry-run)")
				return

			resolve_config = args['resolve_config']
			human_times = args['human_times']

			with open(args['out'], 'w') as f:

				if not cc:
					f.write("[]")
					return

				f.write(
					"[\n" + jsondump(
						self.morph_ret( # type: ignore[operator]
							ctx, cc[0], resolve_config, human_times, id_mapper # type: ignore[index]
						)
					)
				)

				for el in cc[1:]: # type: ignore[index]
					f.write(
						",\n" +
						jsondump(
							self.morph_ret(ctx, el, resolve_config, human_times, id_mapper)
						)
					)
				f.write("\n]")

		# reset or soft-reset below
		elif sub_op.endswith('reset'):

			run_id = ctx.new_run_id()
			changed = t2_utils \
				.i_know_what_i_am_doing() \
				.reset_t2s(col=col, cli=True, run_id=run_id, soft = (sub_op == 'soft-reset'), **args)

			print('Number of changed T2 docs: %i' % changed)

			if changed > 0:
				# Add new doc in the 'events' collection
				EventHandler('CLI', ctx.db).register(run_id=run_id, tier=-1)


	def morph_ret(self,
		ctx: AmpelContext, doc: T2Document,
		resolve_config: bool = False,
		human_times: bool = False,
		id_mapper: None | AbsIdMapper = None
	) -> T2Document:

		if resolve_config and doc['config']:
			doc['config'] = ctx.config._config['confid'].get(doc['config'])

		if human_times:
			for el in doc['meta']:
				if 'ts' in el:
					el['ts'] = datetime.utcfromtimestamp(el['ts']).isoformat() # type: ignore[typeddict-item]

		if id_mapper:
			doc['stock'] = id_mapper.to_ext_id(doc['stock']) # type: ignore[arg-type]

		return doc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsProcessorTemplate.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.07.2021
# Last Modified Date:  16.07.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.log.AmpelLogger import AmpelLogger
from ampel.base.decorator import abstractmethod
from ampel.model.UnitModel import UnitModel
from ampel.config.builder.AbsConfigTemplate import AbsConfigTemplate


class AbsProcessorTemplate(AbsConfigTemplate, abstract=True):

	@abstractmethod
	def get_model(self, config: dict[str, Any], logger: AmpelLogger) -> UnitModel:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/t3/QueryMatchModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.09.2018
# Last Modified Date:  16.03.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.types import strict_iterable
from ampel.util.collections import check_seq_inner_type
from ampel.util.docstringutils import gendocstring
from ampel.base.AmpelBaseModel import AmpelBaseModel

@gendocstring
class QueryMatchModel(AmpelBaseModel):
	"""

	Note: If logic parameter is a string or a simple list,
	it is interpreted as if it was 'anyOf' (OR operator)

	.. sourcecode:: python\n

		In []: QueryMatchModel(**{"field":"ab", "logic":'3'})
		Out[]: <QueryMatchModel field='ab' logic={'anyOf': ['3']}>

		In []: QueryMatchModel(**{"field":"ab", "logic":['3', '1', '2']})
		Out[]: <QueryMatchModel field='ab' logic={'anyOf': ['3', '1', '2']}>

		In []: QueryMatchModel(**{"field":"ab", "logic":{'allOf': ['3', '1', '2']}})
		Out[]: <QueryMatchModel field='ab' logic={'allOf': ['3', '1', '2']}>

		In []: QueryMatchModel(**{"field":"ab", "logic":{'anyOf': ['3', '1', '2']}})
		Out[]: <QueryMatchModel field='ab' logic={'anyOf': ['3', '1', '2']}>

		In []: QueryMatchModel(**{"field":"ab", "logic":{'anyOf': [{'allOf': ['1','2']}, '3', '1', '2']}})
		Out[]: <QueryMatchModel field='ab' logic={'anyOf': [{'allOf': ['1', '2']}, '3', '1', '2']}>

		In []: QueryMatchModel(**{"field":"ab", "logic":{'anyOf': [{'allOf': ['1','2']}, '3', {'allOf': ['1','3']}]}})
		Out[]: <QueryMatchModel field='ab' logic={'anyOf': [{'allOf': ['1', '2']}, '3', {'allOf': ['1', '3']}]}>

		In []: QueryMatchModel(**{"field":"ab", "logic":['1', '2', ['1', '2', '3']]})
		Out[]: Unsupported nesting (err 1)

		In []: QueryMatchModel(**{"field":"ab", "logic":{'allOf': ['1', '2', ['1','2']]}})
		Out[]: Unsupported nesting (err 5)

		In []: QueryMatchModel(**{"field":"ab", "logic":{'allOf': ['1', '2'], 'abc': '2'}})
		Out[]: Unsupported dict format {'allOf': ['1', '2'], 'abc': '2'}

		In []: QueryMatchModel(**{"field":"ab", "logic":{'anyOf': [{'anyOf': ['1','2']}, '2']}})
		Out[]: Unsupported nesting (err 3)

		In []: QueryMatchModel(**{"field":"ab", "logic":{'anyOf': [{'allOf': ['1','2']}, '3', {'anyOf': ['1','2']}]}})
		Out[]: Unsupported nesting (err 3)

		In []: QueryMatchModel(**{"field":"ab", "logic":{'allOf': [{'allOf': ['1','2']}, '3', '1', '2']}})
		Out[]: Unsupported nesting (err 5)

		In []: QueryMatchModel(**{"field":"ab", "logic":{'anyOf': [{'allOf': ['1','2']}, '3', {'allOf': ['1',{'allOf':['1','2']}]}]}})
		Out[]: Unsupported nesting (err 4)
	"""


	field: str
	logic: Any

	def __init__(self, **kwargs):
		kwargs["logic"] = self.check_format(kwargs.get("logic"))
		super().__init__(**kwargs)

	def check_format(self, v: Any) -> dict[str, Any]:

		#print("--------------------------------")
		#print("QueryMatchModel: v: %s" % str(v))
		#print("QueryMatchModel: kwargs: %s" % kwargs)
		#print("QueryMatchModel: values: %s" % values)

		if type(v) is str:
			return {'anyOf': [v]}

		if type(v) is list:
			if not check_seq_inner_type(v, str):
				raise ValueError(
					"QueryMatchModel error\n" +
					"Unsupported nesting (err 1)"
				)
			return {'anyOf': v}

		if type(v) is dict:

			if len(v) != 1:
				raise ValueError(
					"QueryMatchModel error\n" +
					"Unsupported dict format %s" % v
				)

			if 'anyOf' in v:

				if not isinstance(v['anyOf'], strict_iterable):
					raise ValueError(
						"QueryMatchModel error\n" +
						"Invalid dict value type: %s. Must be a sequence" % type(v['anyOf'])
					)

				# 'anyOf' supports only a list of dicts and str
				if not check_seq_inner_type(v['anyOf'], (str, dict), multi_type=True):
					raise ValueError(
						"QueryMatchModel error\n" +
						"Unsupported nesting (err 2)"
					)

				for el in v['anyOf']:

					if isinstance(el, dict):

						if 'anyOf' in el:
							raise ValueError(
								"QueryMatchModel error\n" +
								"Unsupported nesting (err 3)"
							)

						elif 'allOf' in el:

							# 'allOf' closes nesting
							if not check_seq_inner_type(el['allOf'], str):
								raise ValueError(
									"QueryMatchModel error\n" +
									"Unsupported nesting (err 4)"
								)
						else:
							raise ValueError(
								"QueryMatchModel error\n" +
								"Unsupported dict: %s" % el
							)

			elif 'allOf' in v:

				if not isinstance(v['allOf'], strict_iterable):
					raise ValueError(
						"QueryMatchModel error\n" +
						"Invalid dict value type: %s. Must be a sequence" % type(v['anyOf']))

				# 'allOf' closes nesting
				if not check_seq_inner_type(v['allOf'], str):
					raise ValueError(
						"QueryMatchModel error\n" +
						"Unsupported nesting (err 5)"
					)
			else:
				raise ValueError(
					"QueryMatchModel error\n" +
					"Invalid dict key (only 'anyOf' and 'allOf' allowed)"
				)

		return v



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t2/T2Worker.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                24.05.2019
# Last Modified Date:  28.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import random
from math import ceil
from time import time
from bson import ObjectId
from importlib import import_module
from typing import Union, Any, ClassVar, Literal
from collections.abc import Sequence

from ampel.types import T, OneOrMany, UBson, ubson
from ampel.struct.UnitResult import UnitResult
from ampel.enum.DocumentCode import DocumentCode
from ampel.enum.MetaActionCode import MetaActionCode
from ampel.enum.JournalActionCode import JournalActionCode
from ampel.content.StockDocument import StockDocument
from ampel.content.DataPoint import DataPoint
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.base.BadConfig import BadConfig
from ampel.log import AmpelLogger
from ampel.log.utils import convert_dollars
from ampel.log.utils import report_exception, report_error
from ampel.abstract.AbsStockT2Unit import AbsStockT2Unit
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit
from ampel.abstract.AbsStateT2Unit import AbsStateT2Unit
from ampel.abstract.AbsCustomStateT2Unit import AbsCustomStateT2Unit
from ampel.abstract.AbsTiedT2Unit import AbsTiedT2Unit
from ampel.abstract.AbsTiedPointT2Unit import AbsTiedPointT2Unit
from ampel.abstract.AbsTiedStateT2Unit import AbsTiedStateT2Unit
from ampel.abstract.AbsTiedStockT2Unit import AbsTiedStockT2Unit
from ampel.abstract.AbsTiedCustomStateT2Unit import AbsTiedCustomStateT2Unit, U
from ampel.abstract.AbsWorker import AbsWorker, register_stats
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.UnitModel import UnitModel
from ampel.model.StateT2Dependency import StateT2Dependency
from ampel.mongo.utils import maybe_match_array
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.view.T2DocView import T2DocView

AbsT2 = Union[
	AbsStockT2Unit, AbsPointT2Unit, AbsStateT2Unit, AbsTiedPointT2Unit,
	AbsTiedStateT2Unit, AbsCustomStateT2Unit[T], AbsTiedCustomStateT2Unit[T, U]
]

abs_t2 = (
	AbsStockT2Unit, AbsPointT2Unit, AbsStateT2Unit, AbsTiedPointT2Unit,
	AbsTiedStateT2Unit, AbsCustomStateT2Unit, AbsTiedCustomStateT2Unit
)

stat_latency, stat_count = register_stats(tier=2)

class BackoffConfig(AmpelBaseModel):
	code_match: Sequence[int] = [
		DocumentCode.T2_PENDING_DEPENDENCY, # input t2 doc has not yet resolved
		DocumentCode.T2_MISSING_DEPENDENCY, # input t2 docs not found
		DocumentCode.T2_UNKNOWN_LINK, # input t1 doc not found
		DocumentCode.T2_MISSING_INFO, # t0 docs not found
	]
	base: float = 2
	factor: float = 1
	jitter: bool = True


class T2Worker(AbsWorker[T2Document]):
	"""
	Fill results into pending :class:`T2 documents <ampel.content.T2Document.T2Document>`.
	"""

	#: process only those :class:`T2 documents <ampel.content.T2Document.T2Document>`
	#: with the given :attr:`~ampel.content.T2Document.T2Document.code`
	code_match: OneOrMany[int] = [
		DocumentCode.NEW,
		DocumentCode.RERUN_REQUESTED,
		DocumentCode.T2_NEW_PRIO,
		DocumentCode.T2_PENDING_DEPENDENCY
	]

	#: Process dependencies of each T2 doc if they are not already marked OK.
	#: You likely want to disable this when running multiple workers against a
	#: replicated database in order to reduce the chance that concurrent
	#: updates interfere with the retry mechanism.
	run_dependent_t2s: bool = True
	tier: ClassVar[Literal[2]] = 2

	#: Add an exponentially increasing delay on codes that indicate temporary
	#: failures. Different backoff strategies may be specified for different codes
	backoff_on_retry: None | list[BackoffConfig] = [BackoffConfig()]


	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self._backoff_on_retry = {
			code: backoff
			for backoff in (self.backoff_on_retry or [])
			for code in backoff.code_match
		}


	def process_doc(self,
		doc: T2Document,
		stock_updr: MongoStockUpdater,
		logger: AmpelLogger
	) -> tuple[UBson, int]:

		before_run = time()

		t2_unit = self.get_unit_instance(doc, logger)

		if not isinstance(t2_unit, abs_t2):
			raise ValueError(f"Unsupported unit: {doc['unit']}")

		if (trials := len([el for el in doc['meta'] if el['tier'] == 2])) <= self.max_try:
			ret = self.run_t2_unit(t2_unit, doc, logger, stock_updr)
		else:
			ret = UnitResult(code=DocumentCode.TOO_MANY_TRIALS)

		# Used as timestamp and to compute duration below (using before_run)
		now = time()

		# _id is an ObjectId, but declared as bytes in ampel-interface to avoid
		# an explicit dependency on pymongo
		if doc['meta']: # robustify against manual changes of the db
			stat_latency.labels(doc['unit']).observe(now - doc['meta'][0]['ts'])
		stat_count.labels(doc['unit']).inc()
		self._doc_counter += 1
		body = None
		tag = None
		code = 0

		try:

			# New (channel-less) journal entry for the associated stock document
			trace_id = (
				({'t2worker': self._trace_id} if self._trace_id else {}) |
				({'t2unit': t2_unit._trace_id} if t2_unit._trace_id else {})
			)

			jrec = stock_updr.add_journal_record(
				stock = doc['stock'],
				channel = doc['channel'],
				doc_id = doc['_id'], # type: ignore
				trace_id = trace_id or None,
				unit = doc['unit']
			)

			# New meta entry (code appended later)
			meta = self.gen_meta(
				stock_updr.run_id,
				t2_unit._trace_id,
				round(now - before_run, 3),
				MetaActionCode.BUMP_STOCK_UPD
			)

			# The channel-less activity of this t2 worker
			activity = meta['activity'][0]

			# Unit requested customizations
			if isinstance(ret, UnitResult):

				if ret.adapter:
					if ret.adapter not in self._adapters:
						self._adapters[ret.adapter] = getattr(
							import_module(f"ampel.core.adapter.{ret.adapter}"),
							ret.adapter
						)(context=self.context, run_id=stock_updr.run_id)
					ret = self._adapters[ret.adapter].handle(ret)

				if ret.body:
					body = ret.body
					activity['action'] |= MetaActionCode.ADD_BODY
					jrec['action'] |= JournalActionCode.T2_ADD_BODY

				if ret.tag:
					tag = ret.tag
					activity['action'] |= MetaActionCode.ADD_UNIT_TAG
					activity['tag'] = ret.tag
					jrec['action'] |= JournalActionCode.T2_ADD_TAG

				if ret.code:
					code = ret.code
					activity['action'] |= MetaActionCode.SET_UNIT_CODE
					jrec['action'] |= JournalActionCode.T2_SET_CODE
					# set retry time for missing dependencies, similar to
					# https://github.com/litl/backoff
					if (backoff := self._backoff_on_retry.get(ret.code, None)):
						delay = backoff.factor * (backoff.base ** trials)
						if backoff.jitter:
							delay = random.uniform(0, delay)
						meta['retry_after'] = meta['ts'] + ceil(delay)

				# TODO: check that unit did not use system reserved code
				if code != 0 and code in DocumentCode.__members__.values():
					logger.info(
						f'T2 unit {t2_unit.__class__.__name__} returned document '
						f'code: {code} ({DocumentCode(code).name})',
						extra={'unit': doc['unit'], 'stock': doc['stock']}
					)

				if ret.journal:
					jrec.update(ret.journal) # type: ignore
					activity['action'] |= MetaActionCode.EXTRA_JOURNAL
					jrec['action'] |= JournalActionCode.T2_EXTRA_JOURNAL

			# Unit returned bson-like content
			elif isinstance(ret, ubson):
				body = ret
				activity['action'] |= MetaActionCode.ADD_BODY
				jrec['action'] |= JournalActionCode.T2_ADD_BODY

			# Unsupported object returned by unit
			else:
				code = DocumentCode.ERROR
				self._processing_error(
					logger, doc, None, meta, extra={'ret': ret},
					msg='Invalid content returned by T2 unit'
				)

			meta['code'] = code
			self.commit_update(
				{'_id': doc['_id']}, # type: ignore[typeddict-item]
				meta, logger, body=body, tag=tag, code=code
			)

			# Update stock document
			if len(stock_updr._updates) >= self.updates_buffer_size:
				stock_updr.flush()

		except Exception as e:

			if self.raise_exc:
				raise e

			self._processing_error(
				logger, doc, None, exception=e, msg='An exception occured',
				meta = self.gen_meta(stock_updr.run_id, t2_unit._trace_id, round(now - before_run, 3))
			)

		return body, code


	# NB: spell out union arg to ensure a common context for the TypeVar T
	def load_input_docs(self,
		t2_unit: AbsT2, t2_doc: T2Document, logger: AmpelLogger, stock_updr: MongoStockUpdater
	) -> Union[
		None,
		UnitResult,                                              # Error / missing dependency
		DataPoint,                                               # point t2
		tuple[DataPoint, T2DocView],                             # tied point t2
		tuple[StockDocument],                                    # stock t2
		tuple[T1Document, Sequence[DataPoint]],                  # state t2
		tuple[T],                                                # custom state t2 (T could be LightCurve)
		tuple[T1Document, Sequence[DataPoint], list[T2DocView]], # tied state t2
		tuple[T, list[T2DocView]]                                # tied custom state t2
	]:
		"""
		Fetches documents required by `t2_unit`.

		:param t2_unit: instance to fetch input docs for
		:param t2_doc: document for which inputs are needed
		:param logger: logger
		:param stock_updr: journal update service

		Regarding tied T2s (unit depends on other T2s):
		
		- Dependent T2s will be executed if run_dependent_t2s is True.
		- Note that the ingester ingests point and state t2s before state t2s so that the natural ordering
		  of T2 docs to be processed fits rather well a setup with a single T2processor instance processing all T2s.
		- For state t2s tied with other state t2s, putting the dependent units first in the AlertConsumer
		  directives will make sure these are ingested first and thus processed first by T2Processor.
		"""

		# State bound T2 units require loading of compound doc and datapoints
		if isinstance(
			t2_unit, (
				AbsStateT2Unit,
				AbsCustomStateT2Unit,
				AbsTiedStateT2Unit,
				AbsTiedCustomStateT2Unit
			)
		):
			dps: list[DataPoint] = []
			t1_doc: None | T1Document = next(
				self.col_t1.find(
					{'stock': t2_doc['stock'], 'link': t2_doc['link']}
				), None
			)

			# compound doc must exist (None could mean an ingester bug)
			if t1_doc is None:
				report_error(
					self._ampel_db, msg='T1Document not found', logger=logger,
					info={'id': t2_doc['link'], 'doc': t2_doc}
				)
				return UnitResult(code=DocumentCode.T2_UNKNOWN_LINK)

			# Datarights: suppress channel info (T3 uses instead a
			# 'projection' procedure that should not be necessary here)
			t1_doc.pop('channel') # type: ignore[misc]

			t1_dps_ids = list(t1_doc['dps'])

			# Sort DPS from DB in the same order than referenced by 'dps' from t1 doc
			dps = sorted(
				self.col_t0.find({'id': {'$in': t1_dps_ids}}),
				key = lambda dp: t1_dps_ids.index(dp['id'])
			)

			# Should never happen (only in case of ingestion bug)
			if not dps:
				report_error(
					self._ampel_db, msg='Datapoints not found',
					logger=logger, info={'t1': t1_doc, 't2': t2_doc}
				)
				return UnitResult(code=DocumentCode.T2_MISSING_INFO)

			if len(dps) != len(set(t1_dps_ids)):
				missing = sorted(
					set(t1_dps_ids) - {el['id'] for el in dps},
					key = t1_dps_ids.index
				)
				report_error(
					self._ampel_db, msg='Some datapoints not found',
					logger=logger, info={'t1': t1_doc, 't2': t2_doc, 'missing': missing}
				)
				return UnitResult(code=DocumentCode.T2_MISSING_INFO)

			for dp in dps:
				if 'excl' in dp:
					del dp['excl'] # Optional channel based exclusion

			if isinstance(t2_unit, AbsTiedT2Unit):

				queries: list[dict[str, Any]] = []

				if isinstance(t2_unit, AbsTiedCustomStateT2Unit):
					# A LightCurve instance for example
					custom_state = t2_unit.build(t1_doc, dps)

				for tied_model in t2_unit.t2_dependency:

					d = self.build_tied_t2_query(t2_unit, tied_model, t2_doc)
					if d['link'] is None:

						if isinstance(tied_model, StateT2Dependency) and tied_model.link_override:

							filtr, sort, slc = tied_model.link_override.tools()
							l = dps

							if filtr:
								l = filtr.apply(l)

							# Sort (ex: by body.jd)
							if sort:
								l = sort(l)

							# Slice (ex: first datapoint)
							if slc:
								l = l[slc]

							d['link'] = maybe_match_array([el['id'] for el in l])

							if logger.verbose > 1:
								logger.debug(
									f"link_override result: {d['link']}",
									extra={'recipient': t2_doc['unit'], 'stock': t2_doc['stock']}
								)
						else:
							d['link'] = maybe_match_array(t1_dps_ids)

					queries.append(d)

				qres = self.run_tied_queries(queries, t2_doc, stock_updr, logger)

				# Dependency missing
				if isinstance(qres, UnitResult):
					return qres

				if isinstance(t2_unit, AbsTiedStateT2Unit):
					return t1_doc, dps, qres

				# instance of AbsTiedCustomStateT2Unit
				return custom_state, qres

			else:

				if isinstance(t2_unit, AbsStateT2Unit):
					return t1_doc, dps

				else: # instance of AbsCustomStateT2Unit
					return (t2_unit.build(t1_doc, dps), )

		elif isinstance(t2_unit, AbsStockT2Unit):

			if doc := next(self.col_stock.find({'stock': t2_doc['link']}), None):
				return (doc, )

			report_error(
				self._ampel_db, msg='Stock doc not found',
				logger=logger, info={'doc': t2_doc}
			)
			return UnitResult(code=DocumentCode.T2_UNKNOWN_LINK)

		elif isinstance(t2_unit, (AbsPointT2Unit, AbsTiedPointT2Unit)):

			if doc := next(self.col_t0.find({'id': t2_doc['link']}), None):
				if isinstance(t2_unit, AbsTiedPointT2Unit):

					qres = self.run_tied_queries(
						[
							self.build_tied_t2_query(t2_unit, tied_model, t2_doc)
							for tied_model in t2_unit.t2_dependency
						],
						t2_doc, stock_updr, logger
					)

					if isinstance(qres, UnitResult):
						return qres

					return (doc, qres)

				return (doc, )

			report_error(
				self._ampel_db, msg='Datapoint not found',
				logger=logger, info={'doc': t2_doc}
			)
			return UnitResult(code=DocumentCode.T2_UNKNOWN_LINK)

		else:

			report_error(
				self._ampel_db, msg='Unknown T2 unit type',
				logger=logger, info={'doc': t2_doc}
			)

		return None


	def run_tied_queries(self,
		queries: list[dict[str, Any]],
		t2_doc: T2Document,
		stock_updr: MongoStockUpdater,
		logger: AmpelLogger
	) -> UnitResult | list[T2DocView]:

		t2_views: list[T2DocView] = []

		for query in queries:

			if self.run_dependent_t2s:

				processed_ids: list[ObjectId] = []

				# run pending dependencies
				while (
					dep_t2_doc := self.col.find_one_and_update(
						{
							'code': self.code_match if isinstance(self.code_match, DocumentCode)
							else maybe_match_array(self.code_match) # type: ignore[arg-type]
						} | query,
						{'$set': {'code': DocumentCode.RUNNING}}
					)
				):

					if logger.verbose > 1:
						logger.debug(
							'Processing tied t2 docs',
							extra={'unit': dep_t2_doc['unit'], 'stock': dep_t2_doc['stock']}
						)

					if not dep_t2_doc.get('body'):
						dep_t2_doc['body'] = []

					body, code = self.process_doc(dep_t2_doc, stock_updr, logger)
					dep_t2_doc['body'].append(body)
					dep_t2_doc['meta'].append({'code': code, 'tier': 2})

					# suppress channel info
					dep_t2_doc.pop('channel')
					t2_views.append(T2DocView.of(dep_t2_doc, self.context.config))
					processed_ids.append(dep_t2_doc['_id'])

				if len(processed_ids) > 0:
					query['_id'] = {'$nin': processed_ids}

			if logger.verbose > 1:
				logger.debug(
					'Running tied t2 query',
					extra={
						'unit': t2_doc['unit'],
						'stock': t2_doc['stock'],
						'query': convert_dollars(query)
					}
				)

			# collect dependencies
			for dep_t2_doc in self.col.find(query):
				# suppress channel info
				dep_t2_doc.pop('channel')
				t2_views.append(T2DocView.of(dep_t2_doc, self.context.config))

			for view in t2_views:
				if view.get_payload() is None:
					if logger.verbose > 1:
						logger.debug(
							'Dependent T2 unit not run yet',
							extra={
								'unit': view.unit,
								'stock': view.stock,
								'link': view.link,
								't2_type': view.t2_type,
							}
						)
					return UnitResult(code=DocumentCode.T2_PENDING_DEPENDENCY)

		if not t2_views:
			return UnitResult(code=DocumentCode.T2_MISSING_DEPENDENCY)

		return t2_views


	def build_tied_t2_query(self,
		t2_unit: AbsT2, tied_model: UnitModel, t2_doc: T2Document
	) -> dict[str, Any]:
		"""
		This method handles 'default' situations.
		Callers must check returned 'link'.
		If None (state t2 linked with point t2), further handling is required

		:raises:
			- ValueError if functionality is not implemented yet
			- BadConfig if what's requested is not possible (a point T2 cannot be linked with a state t2)
	
		:returns: link value to match or None if further handling is required (state t2 linked with point t2)
		"""

		t2_unit_info = self.context.config.get(
			f'unit.{tied_model.unit}', dict
		)

		if not t2_unit_info:
			raise ValueError(f'Unknown T2 unit {tied_model.unit}')

		d: dict[str, Any] = {
			'unit': tied_model.unit,
			'config': tied_model.config,
			'channel': {'$in': t2_doc['channel']},
			'stock': t2_doc['stock'],
		}

		if isinstance(t2_unit, AbsTiedPointT2Unit):

			if 'AbsPointT2Unit' in t2_unit_info['base']:
				d['link'] = t2_doc['link']

			elif 'AbsStockT2Unit' in t2_unit_info['base']:
				d['link'] = t2_doc['stock']

			else: # State T2
				raise BadConfig('Tied point T2 cannot be linked with state T2s')

		elif isinstance(t2_unit, AbsTiedStockT2Unit):

			if 'AbsPointT2Unit' in t2_unit_info['base']:
				raise BadConfig('Tied stock T2 cannot be linked with point T2s')

			elif 'AbsStockT2Unit' in t2_unit_info['base']:
				d['link'] = t2_doc['stock']

			else: # State T2
				raise BadConfig('Tied stock T2 cannot be linked with state T2s')

		if isinstance(t2_unit, (AbsTiedStateT2Unit, AbsTiedCustomStateT2Unit)):

			if 'AbsPointT2Unit' in t2_unit_info['base']:
				d['link'] = None # Further checks required (link_override check)

			elif 'AbsStockT2Unit' in t2_unit_info['base']:
				d['link'] = t2_doc['stock']

			else: # State T2
				d['link'] = t2_doc['link']

		return d


	def run_t2_unit(self,
		t2_unit: AbsT2, t2_doc: T2Document, logger: AmpelLogger, stock_updr: MongoStockUpdater,
	) -> UBson | UnitResult:
		"""
		Regarding the possible int return code:
		usually, if an int is returned, it should be a DocumentCode member
		but let's not be too restrictive here
		"""

		args: Any = self.load_input_docs(t2_unit, t2_doc, logger, stock_updr)
		if args is None:
			logger.error("Unable to load information required to run t2 unit")
			return UnitResult(code=DocumentCode.INTERNAL_ERROR)

		if isinstance(args, UnitResult):
			return args

		try:

			ret = t2_unit.process(*args)

			if t2_unit._buf_hdlr.buffer: # type: ignore[union-attr]
				t2_unit._buf_hdlr.forward( # type: ignore[union-attr]
					logger, stock=t2_doc['stock']
				)

			return ret

		except Exception as e:

			if self.raise_exc:
				raise e

			# Record any uncaught exceptions in troubles collection.
			report_exception(
				self._ampel_db, logger, exc=e, info={
					'_id': t2_doc['_id'], # type: ignore
					'unit': t2_doc['unit'],
					'config': t2_doc['config'],
					'stock': t2_doc['stock'],
					'link': t2_doc['link'],
					'channel': t2_doc['channel']
				}
			)

			return UnitResult(code=DocumentCode.EXCEPTION)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsT3Stager.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                04.01.2020
# Last Modified Date:  10.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator

from ampel.view.T3Store import T3Store
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.content.T3Document import T3Document
from ampel.core.ContextUnit import ContextUnit


class AbsT3Stager(ContextUnit, AmpelABC, abstract=True):
	""" Supply stock views to one or more T3 units.  """

	#: number of buffers to process at once. Set to 0 to disable chunking
	chunk_size: int = 1000

	#: If > 0, cast ampel buffers into views for each t3 unit (meaning possibly redundantly)
	#: since there is no real read-only struct in python
	paranoia_level: int = 0


	@abstractmethod
	def stage(self,
		data: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> None | Generator[T3Document, None, None]:
		""" Process a chunk of AmpelBuffer instances """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/ProcessMorpher.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  16.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import json
from typing import Any
from importlib import import_module
from ampel.log.AmpelLogger import AmpelLogger, VERBOSE
from ampel.model.ProcessModel import ProcessModel
from ampel.util.recursion import walk_and_process_dict
from ampel.util.mappings import dictify
from ampel.util.hash import build_unsafe_dict_id
from ampel.base.LogicalUnit import LogicalUnit


class ProcessMorpher:
	""" Applies various transformations to process dicts """

	def __init__(self,
		process: dict[str, Any],
		templates: dict[str, Any],
		logger: AmpelLogger,
		verbose: bool = False,
		deep_copy: bool = False
	) -> None:

		self.process = json.loads(json.dumps(process)) if deep_copy else process
		self.templates = templates
		self.logger = logger
		self.verbose = verbose


	def get(self) -> dict[str, Any]:
		""" :raises: Error if the morphed process does not comply with ProcessModel """
		return ProcessModel(**self.process).dict()


	def generate_version(self, first_pass_config: dict) -> 'ProcessMorpher':
		"""
		Generates a version identifier for the underyling process as required by ProcessModel

		:param first_pass_config: required to get the version identifier of the ampel units used by this process
		"""
		
		# TODO: implement
		if not self.process.get("version"):
			self.process['version'] = build_unsafe_dict_id(
				dictify(self.process['processor'].get("config", {}))
			)

		return self


	def enforce_t3_channel_selection(self, chan_name: str) -> 'ProcessMorpher':

		if self.process['tier'] == 3:

			for block in self.process['processor']['config']['execute']:
				if block["unit"] != "T3ReviewUnitExecutor":
					raise ValueError(
						f'Cannot enforce channel selection: '
						f'unknown executor {block["unit"]}'
					)
				directive = block['config']['supply']['config']
				if 'select' in directive:
					select = directive['select']
					if select['unit'] not in ('T3StockSelector', 'T3FilteringStockSelector'):
						raise ValueError(
							f'Cannot enforce channel selection: '
							f'unknown stock selection unit {select["unit"]}'
						)
				else:
					select = {'unit': 'T3StockSelector', 'config': {}}
					directive['select'] = select

				if 'config' not in select:
					select['config'] = {}

				if self.verbose:
					action = 'Modifying' if 'channel' in select['config'] else 'Adding'
					self.logger.log(VERBOSE,
						f'{action} channel selection criteria ({chan_name}) '
						f'for process {self.process["name"]}'
					)

				# processes embedded in channel must feature a transient selection.
				# An exception will be raised if someone embeds an admin t3 proc (without selection)
				# within a channel (unsupported feature)
				select['config']['channel'] = chan_name

		return self


	def apply_template(self, first_pass_config: dict) -> 'ProcessMorpher':
		""" Applies template possibly associated with process """

		# The process embedded in channel def requires templating itself
		if 'template' in self.process:

			if self.verbose:
				self.logger.log(VERBOSE,
					f'Applying template {self.process["template"]} '
					f'to process {self.process["name"]} '
					f'from distribution {self.process["distrib"]}'
				)

			self.process = self.templates[
				self.process['template']
			](**self.process).get_process(first_pass_config, self.logger)

		return self


	def scope_aliases(self, first_pass_config: dict) -> 'ProcessMorpher':
		""" Note: should be called before hash_t2_config """

		if self.process.get('distrib'):

			if self.verbose:
				self.logger.debug("Scoping aliases")

			recurse = False

			while walk_and_process_dict(
				arg = self.process,
				callback = self._scope_alias_callback,
				match = ['config'],
				first_pass_config = first_pass_config
			):
				if self.verbose:
					self.logger.debug("Alias(es) scoped %s" % ("again" if recurse else ""))
				recurse = True
		else:

			self.logger.error(
				f'Cannot scope aliases for process {self.process["name"]}'
				f' as it is missing a distribution name'
			)

		return self


	def _scope_alias_callback(self, path, k, d, **kwargs) -> bool:
		"""
		:returns: true if modification was done
		"""

		if 'first_pass_config' not in kwargs:
			raise ValueError('Parameter "first_pass_config" missing in kwargs')

		v = d[k]

		if v and isinstance(v, str):

			# Global alias
			if v[0] == '%':
				return False

			# Alias already scoped
			if "/" in v:
				return False

			scoped_alias = f'{self.process["distrib"]}/{v}'

			if not any([
				scoped_alias in kwargs['first_pass_config']['alias'][f't{tier}']
				for tier in (0, 1, 2, 3)
			]):
				raise ValueError(f'Alias "{scoped_alias}" not found')

			# Overwrite
			d[k] = scoped_alias

			if self.verbose:
				self.logger.log(VERBOSE, f'Config alias "{v}" renamed into "{scoped_alias}"')

			return True

		return False


	def resolve_aliases(self, t2d: dict[str, Any], aliases: dict[str, Any], root_path: str):
		"""
		Resolves aliases recursively (necessary before hashing t2 config)
		"""

		if self.verbose:
			self.logger.debug("Resolving aliases (required for hash)")

		recurse = False

		while walk_and_process_dict(
			arg = t2d,
			callback = self._resolve_alias_callback,
			match = ['config'],
			aliases = aliases,
			root_path = root_path
		):
			if self.verbose:
				self.logger.debug("Alias(es) resolved %s" % ("again" if recurse else ""))
			recurse = True


	def _resolve_alias_callback(self, path, k, d, **kwargs) -> None:
		"""
		Used by walk_and_process_dict(...) from resolve_aliases(...)
		"""

		if not isinstance(d[k], str):
			return

		aliases = kwargs.get('aliases')

		if not aliases:
			raise ValueError('Parameter "aliases" missing in kwargs')

		if d[k] not in aliases:
			raise ValueError(
				f'Unknown T2 config alias ({d[k]}) defined in process {self.process["name"]}'
			)

		if self.verbose:
			self.logger.debug(f"Resolving alias '{d[k]}' in {kwargs.get('root_path')}")

		d[k] = aliases[d[k]]


	def hash_t2_config(self, out_config: dict) -> 'ProcessMorpher':
		"""
		This method modifies the underlying self._process dict structure.
		The 'config' (dict) value of UnitModel instances is replaced by a hash value (int).
		Notes:
		- Applies only to t0 and t1 processes
		- The config path filter "t2_compute.units" is used
		- Aliases are resolved (works recursively so that t2_dependency of tied units can use aliases as well)
		
		:param out_config: used to store new map entries in the ampel config: {<confid>: {<hash>: <conf dict>}}.
		"""

		if self.process['tier'] not in (0, 1):
			return self

		if self.verbose:
			self.logger.debug("Looking for t2 config to hash")

		# Example of conf_dicts keys
		# processor.config.directives.0.combine.0.state_t2.0
 		# processor.config.directives.0.combine.0.state_t2.0.config.t2_dependency.0
 		# processor.config.directives.0.point_t2.0
		conf_dicts: dict[str, dict[str, Any]] = {}

		walk_and_process_dict(
			arg = self.process["processor"]["config"]["directives"],
			callback = self._gather_t2_config_callback,
			match = ['point_t2', 'stock_t2', 'state_t2'],
			conf_dicts = conf_dicts
		)

		# This does the trick of processing nested config first
		sorted_conf_dicts = {
			k: conf_dicts[k]
			for k in sorted(conf_dicts, key=len, reverse=True)
		}

		for k, d in sorted_conf_dicts.items():

			t2_unit = d["unit"]
			conf = d.get("config", {})

			if self.verbose:
				extra = {'process': self.process['name'], 'conf': k + ".config"}
				self.logger.debug("Hashing T2 config", extra=extra)

			# Replace alias with content
			if isinstance(conf, str):
			
				if conf not in out_config['alias']['t2']:
					raise ValueError(
						f'Unknown T2 config alias ({conf}) defined in process {self.process["name"]}'
					)

				if self.verbose:
					self.logger.debug(f"Resolving alias '{conf}'", extra=extra)

				conf = out_config['alias']['t2'][conf]

			if isinstance(conf, dict):

				if override := d.get("override"):
					conf |= override

				if fqn := out_config['unit'][t2_unit].get('fqn'):
					T2Unit = getattr(import_module(fqn), fqn.split('.')[-1])
					conf = T2Unit.validate(conf) # dictify ?
				else:
					self.logger.warn(
						f"T2 unit {t2_unit} not installed locally. "
						f"Building *unsafe* conf dict hash: "
						f"changes in unit defaults between releases will go undetected",
						extra=extra
					)

				if self.verbose:
					self.logger.debug("Computing hash", extra=extra)

				if conf is None:
					d["config"] = None
				else:
					d["config"] = out_config['confid'].add(conf, None, None, None)

			# For internal use only
			elif isinstance(conf, int):
				if conf not in out_config['confid']:
					raise ValueError(
						f'Unknown T2 config (int) alias defined by {k}:\n {t2_unit}'
					)
			else:
				raise ValueError(
					f'Unknown T2 config defined by {k}:\n {t2_unit}'
				)

		return self


	def _gather_t2_config_callback(self, path, k, d, **kwargs) -> None:
		"""
		Used by walk_and_process_dict(...) from hash_t2_config(...)
		"""

		if self.verbose:
			self.logger.info(f"# path: {path}.{k}")

		if d[k]:
			for i, el in enumerate(d[k]):
				if self.verbose:
					self.logger.info(f"# path: {path}.{k}.{i}")
					self.logger.info(el)
				kwargs['conf_dicts'][f"{path}.{k}.{i}"] = el



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/supply/complement/T3RandIntAppender.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.06.2020
# Last Modified Date:  14.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from random import randint
from typing import Iterable
from ampel.view.T3Store import T3Store
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.abstract.AbsBufferComplement import AbsBufferComplement


class T3RandIntAppender(AbsBufferComplement):
	"""
	Subclass of ContextUnit because implementing classes might need access to
	the AmpelConfig (foremost to the contained resource definitions)
	"""

	int_from: int = 0
	int_to: int = 100

	def complement(self, it: Iterable[AmpelBuffer], t3s: T3Store) -> None:

		for b in it:

			if b.get('extra', None):
				if isinstance(b['extra'], dict):
					if 'randint' in b['extra']:
						self.logger.error("Randint already defined in ampel buffer", extra=b) # type: ignore[arg-type]
				else:
					raise ValueError("AmpelBuffer 'extra' should be a dict")
			else:
				b['extra'] = {}

			b['extra']['randint'] = randint(self.int_from, self.int_to)  # type: ignore[index]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/t3/AliasableModel.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  Unspecified
# Last Modified By:    jvs

from typing import Any, ClassVar, TYPE_CHECKING
from ampel.base.AmpelBaseModel import AmpelBaseModel

if TYPE_CHECKING:
	from ampel.config.AmpelConfig import AmpelConfig


class AliasableModel(AmpelBaseModel):
	"""
	A model that can be initialized from a global alias in the alias.t3 section
	of an AmpelConfig
	"""

	_config: ClassVar['None | AmpelConfig'] = None

	@classmethod
	def validate(cls, value: Any) -> "AliasableModel":
		if cls._config and isinstance(value, str):
			d = cls._config.get(f"alias.t3.%{value}", dict)
			if d:
				value = d
			else:
				raise ValueError(f"{cls.__name__} alias '{value}' not found in Ampel config")
		return super().validate(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-interface/ampel/util/logicschema.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.04.2021
# Last Modified Date:  03.04.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Union, Literal, Any
from collections.abc import Sequence
from ampel.types import T, strict_iterable
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.util.collections import check_seq_inner_type


def to_logical_dict(v, field_name: int | str | dict[str, Any]) -> dict[str, Any]:
	"""
	Converts str/int into {'any_of': int/str}.
	Checks structure in case dict is provided
	(how deep potential nested logic are, raises error in case of incorrect logic).
	:raises ValueError: in case provided argument is incorrect
	"""

	if isinstance(v, list):
		raise ValueError(
			"stock->select->%s config error\n" % field_name +
			"'%s' parameter cannot be a list. " % field_name +
			"Please use the following syntax:\n" +
			" -> {'any_of': ['Ab', 'Cd']} or\n" +
			" -> {'all_of': ['Ab', 'Cd']} or\n" +
			" -> {'one_of': ['Ab', 'Cd']} or\n" +
			"One level nesting is allowed, please see\n" +
			"conditional_expr_converter(..) docstring for more info"
		)

	if isinstance(v, (str, int)):
		return {'any_of': [v]}

	if isinstance(v, dict):

		if len(v) != 1:
			raise ValueError(
				"stock->select->%s config error\n" % field_name +
				"Unsupported dict format %s" % v
			)

		if 'any_of' in v:

			if not isinstance(v['any_of'], strict_iterable):
				raise ValueError(
					"stock->select->%s:any_of config error\n" % field_name +
					"Invalid dict value type: %s. Must be a sequence" % type(v['any_of'])
				)

			# 'any_of' supports only a list of dicts and str/int
			if not check_seq_inner_type(v['any_of'], (str, int, dict), multi_type=True):
				raise ValueError(
					"stock->select->%s:any_of config error\n" % field_name +
					"Unsupported nesting (err 2)"
				)

			if not check_seq_inner_type(v['any_of'], (int, str)) and len(v['any_of']) < 2:
				raise ValueError(
					"stock->select->%s:any_of config error\n" % field_name +
					"any_of list must contain more than one element when containing all_of\n" +
					"Offending value: %s" % v
				)

			for el in v['any_of']:

				if isinstance(el, dict):

					if 'any_of' in el:
						raise ValueError(
							"stock->select->%s:any_of.any_of config error\n" % field_name +
							"Unsupported nesting (any_of in any_of)"
						)

					if 'all_of' in el:

						# 'all_of' closes nesting
						if not check_seq_inner_type(el['all_of'], (int, str)):
							raise ValueError(
								"stock->select->%s:any_of.all_of config error\n" % field_name +
								"Unsupported nesting (all_of list content must be str/int)"
							)

						if len(set(el['all_of'])) < 2:
							raise ValueError(
								"stock->select->%s:all_of config error\n" % field_name +
								"Please do not use all_of with just one element\n" +
								"Offending value: %s" % el
							)

					else:
						raise ValueError(
							"stock->select->%s:any_of config error\n" % field_name +
							"Unsupported nested dict: %s" % el
						)

		elif 'all_of' in v:

			# 'all_of' closes nesting
			if (
				not isinstance(v['all_of'], strict_iterable) or
				not check_seq_inner_type(v['all_of'], (int, str))
			):
				raise ValueError(
					"stock->select->%s:all_of config error\n" % field_name +
					"Invalid type for value %s\n(must be a sequence, is: %s)\n" %
					(v['all_of'], type(v['all_of'])) +
					"Note: no nesting is allowed below 'all_of'"
				)

			if len(set(v['all_of'])) < 2:
				raise ValueError(
					"stock->select->%s:all_of config error\n" % field_name +
					"Please do not use all_of with just one element\n" +
					"Offending value: %s" % v
				)

		elif 'one_of' in v:

			# 'one_of' closes nesting
			if (
				not isinstance(v['one_of'], strict_iterable) or
				not check_seq_inner_type(v['one_of'], (int, str))
			):
				raise ValueError(
					"stock->select->%s:one_of config error\n" % field_name +
					"Invalid type for value %s\n(must be a sequence, is: %s)\n" %
					(v['one_of'], type(v['one_of'])) +
					"Note: no nesting is allowed below 'one_of'"
				)

		else:
			raise ValueError(
				"stock->select->%s config error\n" % field_name +
				"Invalid dict key (only 'any_of', 'all_of', 'one_of' are allowed)"
			)

	return v


def reduce_to_set(
	arg: Union[
		T,
		# unsure if mypy understands unions of dicts with different key literals actually
		dict[Union[Literal['all_of', 'one_of']], Sequence[T]],
		dict[Literal['any_of'], Union[Sequence[T], dict[Literal['all_of'], Sequence[T]]]],
		AllOf[T], AnyOf[T], OneOf[T]
	],
	in_type: tuple[type, ...] = (str, int)
) -> set[T]:
	"""
	.. sourcecode:: python\n
		for schema in (a,b,c,d,e):
			print("Schema: %s" % schema)
			print("Reduced set: %s" % reduce_to_set(schema))

		Schema: 'a'
		Reduced set: {'a'}
		Schema: {'any_of': ['a', 'b', 'c']}
		Reduced set: {'b', 'a', 'c'}
		Schema: {'all_of': [1, 2, 3]}
		Reduced set: {1, 2, 3}
		Schema: {'any_of': [{'all_of': ['a', 'b']}, 'c']}
		Reduced set: {'b', 'a', 'c'}
		Schema: {'any_of': [{'all_of': ['a', 'b']}, {'all_of': ['a', 'c']}, 'd']}
		Reduced set: {'d', 'b', 'a', 'c'}
	"""

	if isinstance(arg, in_type):
		return {arg} # type: ignore

	elif isinstance(arg, (AllOf, AnyOf, OneOf)):
		v: Any = arg.dict()

	elif isinstance(arg, dict):
		v = arg

	else:
		raise ValueError("Unsupported arg type: %s" % type(arg))

	if "any_of" in v:
		s = set()
		for el in v['any_of']:
			if isinstance(el, in_type):
				s.add(el)
			elif isinstance(el, dict):
				for ell in next(iter(el.values())):
					s.add(ell)
			else:
				raise ValueError("unsupported format (1)")
		return s

	if 'all_of' in v:
		return set(v['all_of'])

	if 'one_of' in v:
		return set(v['one_of'])

	raise ValueError("unsupported format (2)")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/BaseViewGenerator.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                20.04.2021
# Last Modified Date:  05.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import TypeVar
from collections.abc import Generator
from ampel.types import StockId, T3Send
from ampel.view.SnapView import SnapView
from ampel.struct.JournalAttributes import JournalAttributes
from ampel.struct.StockAttributes import StockAttributes
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater

T = TypeVar("T", bound=SnapView)

class BaseViewGenerator(Generator[T, T3Send, None]):
	""" Meant to be subclassed """

	def __init__(self, unit_name: str, stock_updr: MongoStockUpdater) -> None:
		self.unit_name = unit_name
		self.stock_updr = stock_updr
		self.stocks: list[StockId] = []

	def send(self, jt: T3Send):

		if isinstance(jt, tuple):
			stock_id = jt[0]
			element: JournalAttributes | StockAttributes = jt[1]
		else:
			stock_id, element = self.stocks[-1], jt

		if isinstance(element, StockAttributes):
			tag = element.tag
			name = element.name
			jattrs = element.journal
		else:
			tag, name = None, None
			jattrs = element

		if tag:
			self.stock_updr.add_tag(stock_id, tag)

		if name:
			self.stock_updr.add_name(stock_id, name)

		self.stock_updr.add_journal_record(stock=stock_id, jattrs=jattrs, unit=self.unit_name)

	def throw(self, type=None, value=None, traceback=None):
		raise StopIteration

	def get_stock_ids(self):
		return self.stocks



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/ingest/T2Compiler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                01.01.2018
# Last Modified Date:  21.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.types import ChannelId, UnitId, T2Link, StockId
from ampel.content.T2Document import T2Document
from ampel.content.MetaActivity import MetaActivity
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.abstract.AbsCompiler import AbsCompiler, ActivityRegister


class T2Compiler(AbsCompiler):
	"""
	Helps build a minimal set of :class:`T2Document <ampel.content.T2Document.T2Document>`
	Multiple T2 documents associated with the same stock record accross different channels
	are merged into one single T2 document that references all corresponding channels.
	"""

	col: None | str = None
	t2s: dict[
		# key: (unit name, unit config, link, stock)
		tuple[UnitId, None | int, T2Link, StockId],
		tuple[
			set[ChannelId], # channels (doc)
			dict[
				frozenset[tuple[str, Any]], # key: traceid
				tuple[ActivityRegister, dict[str, Any]] # activity register, meta_extra
			]
		]
	] = {}


	def add(self, # type: ignore[override]
		unit: UnitId,
		config: None | int,
		stock: StockId,
		link: T2Link,
		channel: ChannelId,
		traceid: dict[str, Any],
		activity: None | MetaActivity | list[MetaActivity] = None,
		meta_extra: None | dict[str, Any] = None
	) -> None:
		"""
		:param tag: tag(s) to be added to T2Document. A corresponding dedicated channel-less
		meta entry will be created (with actin MetaActionCode.ADD_INGEST_TAG and/or ADD_OTHER_TAG)
		"""

		# Doc id
		k = (unit, config, link, stock)
		tid = frozenset(traceid.items())

		# a: tuple({channels}, dict[traceid, (activity register, meta_extra)])
		if a := self.t2s.get(k):

			a[0].add(channel) # set of all channels

			# One meta record will be created for each unique trace id
			# (the present compiler could be used by multiple handlers configured differently)
			if tid in a[1]:

				# (activity register, meta_extra)
				b = a[1][tid]

				# update internal register
				self.register_meta_info(b[0], b[1], channel, activity, meta_extra)

			else:
				a[1][tid] = self.new_meta_info(channel, activity, meta_extra)
		else:
			self.t2s[k] = {channel}, {tid: self.new_meta_info(channel, activity, meta_extra)}


	def commit(self, ingester: AbsDocIngester[T2Document], now: int | float, **kwargs) -> None:

		for k, v in self.t2s.items():

			# Note: mongodb maintains key order
			d: T2Document = {'unit': k[0], 'config': k[1], 'link': k[2]} # type: ignore[typeddict-item]

			if self.col:
				d['col'] = self.col

			d['stock'] = k[3]

			if self.origin:
				d['origin'] = self.origin

			d['channel'] = list(v[0])
			d['meta'], tags = self.build_meta(v[1], now)

			if tags:
				d['tag'] = tags

			ingester.ingest(d)

		self.t2s.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/AbsForwardConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  02.03.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Type
from collections.abc import Sequence
import ampel.config.builder.FirstPassConfig as fpc # avoid circular import issue
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelABC import AmpelABC
from ampel.log.AmpelLogger import AmpelLogger, VERBOSE
from ampel.config.builder.DisplayOptions import DisplayOptions


class AbsForwardConfigCollector(dict, AmpelABC, abstract=True):


	def __init__(self,
		# Forward reference type hint to avoid cyclic import issues
		root_config: 'fpc.FirstPassConfig',
		conf_section: str,
		options: DisplayOptions,
		target_collector_type: type,
		logger: None | AmpelLogger = None
	) -> None:

		self.has_error = False
		self.verbose = options.verbose
		self.root_config = root_config
		self.conf_section = conf_section
		self.target_collector_type = target_collector_type
		self.logger = AmpelLogger.get_logger() if logger is None else logger

		if self.verbose:
			self.logger.log(VERBOSE,
				f'Creating {self.__class__.__name__} collector '
				f'for config section "{conf_section}"'
			)


	def add(self,
		arg: dict[str, Any] | list[Any] | str,
		dist_name: str,
		version: str | float | int,
		register_file: str,
	) -> None:


		for el in [arg] if isinstance(arg, (dict, str)) else arg:

			path_elements = self.get_path(el, dist_name, version, register_file)
			if not path_elements:
				self.error(f' Follow-up error: could not identify routing for {el}')
				continue

			d = self.root_config
			for path in path_elements:
				d = d[path]

			if not isinstance(d, self.target_collector_type):
				self.error(
					f'Routing destination must be an instance of {self.target_collector_type}'
				)
				self.error(
					f'Type of config element with path '
					f'{".".join(str(x) for x in path_elements)}: {type(d)}'
				)
				return

			d.add(el, dist_name, version, register_file) # type: ignore


	@abstractmethod
	def get_path(self,
		arg: dict[str, Any] | str,
		dist_name: str,
		version: str | float | int,
		register_file: str,
	) -> None | Sequence[int | str]:
		...


	def error(self, msg: str, exc_info: None | Any = None) -> None:
		self.logger.error(msg, exc_info=exc_info)
		self.has_error = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/ingest/DualIngestDirective.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                20.05.2021
# Last Modified Date:  21.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Literal
from ampel.types import ChannelId
from ampel.model.ingest.IngestBody import IngestBody
from ampel.model.ingest.FilterModel import FilterModel
from ampel.base.AmpelBaseModel import AmpelBaseModel


class DualIngestDirective(AmpelBaseModel):
	"""
    channel
    filter
    ingest
      new               <- stock was never previously associated with the current channel
        stock_t2
        point_t2        <- based on input dps list
        combine         <- same
          state_t2      <- based on dps list returned by combine
          point_t2      <- same
        mux             <- based on input dps list
          insert
            point_t2    <- based on <datapoints insert> result from muxer unit (first list)
          combine
            state_t2    <- based on <datapoints combine> result from muxer unit (second list)
            point_t2    <- same
      known             <- stock was previously associated with the current channel
        stock_t2
        point_t2        <- based on input dps list
        combine         <- same
          state_t2      <- based on dps list returned by combine
          point_t2      <- same
        mux             <- based on input dps list
          insert
            point_t2    <- based on <datapoints insert> result from muxer unit (first list)
          combine
            state_t2    <- based on <datapoints combine> result from muxer unit (second list)
            point_t2    <- same
	"""

	#: Channel for which to create documents
	channel: ChannelId

	#: Potientially filter input datapoints
	filter: None | FilterModel = None

	ingest: dict[Literal['new', 'known'], IngestBody]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/project/T3ChannelProjector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                07.01.2020
# Last Modified Date:  22.11.2020
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

from typing import Any
from collections.abc import Sequence
from ampel.types import ChannelId
from ampel.log import VERBOSE
from ampel.util.logicschema import reduce_to_set
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.aux.ComboDictModifier import ComboDictModifier
from ampel.t3.stage.project.T3BaseProjector import T3BaseProjector


class T3ChannelProjector(T3BaseProjector):

	channel: ChannelId | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId]

	# Whether to cast structures into immutables objects after modification
	freeze: bool = True

	# Whether fields from input ampel_buffer can be directly altered/modified
	# This is the case if the parent T3UnitRunner contains only one run block
	# If this is not the case, new dict instances must be created to modify existing dicts
	unalterable: bool = True


	def __init__(self, **kwargs) -> None:
		"""
		:param channel: the channel id of the channel for which projections should be performed
		:param unalterable: see ComboDictModifier docstring
		:param freeze: see ComboDictModifier docstring
		"""

		super().__init__(**kwargs)

		self.verbose = self.logger.verbose
		if self.verbose:
			self.logger.log(VERBOSE, f"Setting up channel project for '{self.channel}'")
		self._channel_set: set[ChannelId] = reduce_to_set(self.channel)

		journal_modifier = ComboDictModifier(
			logger = self.logger,
			unalterable = self.unalterable,
			freeze = self.freeze,
			modifications = [
				# Modified ex: {"HU_RANDOM": 3213143434, "HU_RAPID": 43789574389}
				ComboDictModifier.KeepOnlyModel(op="keep_only", key="ts", keep=list(self._channel_set)),
				# Added ex: {"msg": "test", "tier":0, "channel": ["HU_RANDOM", "HU_RAPID"]}
				ComboDictModifier.FuncModifyModel(op="modify", key="journal", func=self.channel_projection),
				ComboDictModifier.FuncModifyModel(op="modify", key="channel", func=self.overwrite_root_channel)
			]
		)

		self.add_func_projector("stock", journal_modifier.apply, first=True)

		for key in ("t1", "t2"):
			self.add_func_projector(key, self.channel_projection, first=True) # type: ignore


	def overwrite_root_channel(self, v: Sequence[ChannelId]) -> None | Sequence[ChannelId]:
		if subset := list(self._channel_set.intersection(v)):
			return subset
		else:
			return None


	def channel_projection(self, dicts: Sequence[dict[str, Any]]) -> Sequence[dict[str, Any]]:
		"""
		Filters out dict entries not associated with configured channel
		Note: debug ouput handled by super class
		"""

		channel_set = self._channel_set
		setitem = dict.__setitem__
		ret: list[dict] = []

		if not dicts:
			return []

		for el in dicts:
			if elchan := el.get('channel'):
				if isinstance(elchan, (str, int)):
					if elchan in channel_set:
						ret.append(el)
				else:
					if subset := list(channel_set.intersection(elchan)):
						channels = (
							list(subset)
							if len(subset) > 1 else list(subset)[0]
						)
						if self.unalterable:
							ret.append({**el, 'channel': channels})
						else:
							setitem(el, 'channel', channels)
							ret.append(el)

		return tuple(ret)



from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry


def test_autoname():
    assert (
        AmpelMetricsRegistry.counter("foo", "foos", unit="blarghs", subsystem="blah")._name
        == "ampel_blah_foo_blarghs"
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/util/collections.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                07.06.2018
# Last Modified Date:  10.09.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from itertools import islice
from collections.abc import Sequence as sequence, Iterable as iterable, Sized as sized
from typing import Any
from collections.abc import Generator, Iterable

from ampel.types import strict_iterable, StrictIterable, T


def ampel_iter(arg: Any) -> Any:
	"""
	-> suppresses python3 treatment of str as iterable (a questionable choice)
	-> Makes None iterable
	"""
	return [arg] if isinstance(arg, (type(None), str, int, bytes, bytearray)) else arg


def get_chunks(seq: Iterable[T], n: int) -> Generator[list[T], None, None]:
	"""
	Yield chunks of length `n` from `seq`

	In []: get_chunks([i for i in range(10)], 2)
	Out[]: <generator object get_chunks at 0x132a26a40>

	In []: list(get_chunks([i for i in range(10)], 2))
	Out[]: [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
	"""

	source = iter(seq)
	while True:
		if chunk := list(islice(source, n)):
			yield chunk
			if len(chunk) < n:
				break
		else:
			break


def get_chunk_sizes(total_size: int, interval_len: int) -> list[int]:
	"""
	In []: get_chunk_sizes(123, 10)
	Out[]: [12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 3]
	"""
	if total_size < interval_len:
		raise ValueError()
	l = [(total_size // interval_len)] * interval_len
	if (total_size % interval_len):
		l.append(total_size % interval_len)
	return l


def try_reduce(arg: Any) -> Any:
	"""
	Returns element contained by sequence if sequence contains only one element.
	Example:
	try_reduce(['ab']) -> returns 'ab'
	try_reduce({'ab'}) -> returns 'ab'
	try_reduce('ab') -> returns 'ab'
	try_reduce(['a', 'b']) -> returns ['a', 'b']
	try_reduce({'a', 'b'}) -> returns {'a', 'b'}
	try_reduce(dict(a=1).keys()) -> returns 'a'
	try_reduce(dict(a=1, b=1).keys()) -> returns dict_keys(['a', 'b'])
	"""

	if isinstance(arg, sized) and len(arg) == 1:
		if isinstance(arg, sequence):
			return arg[0]
		return next(iter(arg)) # type: ignore

	return arg


def to_set(arg) -> set:
	"""
	Reminder of python questionable logic:
	In []: set('abc')
	Out[]: {'a', 'b', 'c'}
	In []: {'abc'}
	Out[]: {'abc'}

	In []: to_set("abc")
	Out[]: {'abc'}
	In []: to_set(["abc"])
	Out[]: {'abc'}
	In []: to_set(['a','b','c'])
	Out[]: {'a', 'b', 'c'}
	In []: to_set([1,2])
	Out[]: {1, 2}
	"""
	return set(arg) if isinstance(arg, strict_iterable) else {arg}


def to_list(arg: int | str | bytes | bytearray | list | Iterable) -> list:
	"""
	raises ValueError is arg is not int, str, bytes, bytearray, list, or Iterable
	"""
	if isinstance(arg, (int, str, bytes, bytearray)):
		return [arg]
	if isinstance(arg, list):
		return arg
	if isinstance(arg, iterable):
		return list(arg)

	raise ValueError(f"Unsupported argument type ({type(arg)})")


def check_seq_inner_type(
	seq, types: type | tuple[type, ...], multi_type: bool = False
) -> bool:
	"""
	check type of all elements contained in a sequence.
	*all* members of the provided sequence must match with:
		* multi_type == False: one of the provided type.
		* multi_type == True: any of the provided type.

	check_seq_inner_type((1,2), int) -> True
	check_seq_inner_type([1,2], int) -> True
	check_seq_inner_type((1,2), float) -> False
	check_seq_inner_type(('a','b'), str) -> True
	check_seq_inner_type((1,2), (int, str)) -> True
	check_seq_inner_type((1,2,'a'), (int, str)) -> False
	check_seq_inner_type((1,2,'a'), (int, str), multi_type=True) -> True

	Note:
	check_seq_inner_type('dsda', str) -> True
	check_seq_inner_type(23, int) -> False
	"""

	# Wrong input
	if not isinstance(seq, sequence) or isinstance(seq, str):
		return False

	# monotype
	if not isinstance(types, sequence):
		return all(isinstance(el, types) for el in seq)

	# different types accepted ('or' connected)
	if multi_type:
		return all(isinstance(el, types) for el in seq)

	return any(
		tuple(check_seq_inner_type(seq, _type) for _type in types)
	)


def has_nested_type(obj: StrictIterable, target_type: type, strict: bool = True) -> bool:
	"""
	:param obj: object instance (dict/list/set/tuple)
	:param target_type: example: ReadOnlyDict/list
	:param strict: must be an instance of the provided type (subclass instances would be rejected)
	"""

	if strict:
		# pylint: disable=unidiomatic-typecheck
		if type(obj) is target_type:
			return True
	else:
		if isinstance(obj, target_type):
			return True

	if isinstance(obj, dict):
		for el in obj.values():
			if has_nested_type(el, target_type):
				return True

	elif isinstance(obj, strict_iterable):
		for el in obj:
			if has_nested_type(el, target_type):
				return True

	return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsOpsUnit.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  Unspecified
# Last Modified By:    jvs

from typing import Any
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.core.ContextUnit import ContextUnit
from ampel.log.AmpelLogger import AmpelLogger
from ampel.types import Traceless

class AbsOpsUnit(AmpelABC, ContextUnit, abstract=True):
    """
    A unit for performing scheduled maintenance tasks not associated with a
    particular processing tier: collecting metrics, reporting exceptions, etc.
    """

    logger: Traceless[AmpelLogger]

    @abstractmethod
    def run(self, beacon: None | dict[str, Any] = None) -> None | dict[str, Any]:
        """
        :param beacon: the result of the previous run
        :returns:
          a BSON-serializable document summarizing the run. This will be
          supplied to the next invocation as `beacon`.
        """
        ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/supply/load/T3SimpleDataLoader.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.12.2019
# Last Modified Date:  29.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from bson.codec_options import CodecOptions
from collections.abc import Iterable, Iterator
from typing import ClassVar
from ampel.types import StockId, StrictIterable
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.abstract.AbsT3Loader import AbsT3Loader
from ampel.mongo.view.FrozenValuesDict import FrozenValuesDict


class T3SimpleDataLoader(AbsT3Loader):
	"""Load all requested documents for the selected stocks"""

	codec_options: ClassVar[None | CodecOptions] = CodecOptions(document_class=FrozenValuesDict)

	def load(self,
		stock_ids: StockId | Iterator[StockId] | StrictIterable[StockId]
	) -> Iterable[AmpelBuffer]:

		return self.data_loader.load(
			stock_ids = stock_ids,
			directives = self.directives,
			channel = self.channel,
			codec_options = self.codec_options,
			logger = self.logger
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/ingest/T1CombineCompute.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                27.05.2021
# Last Modified Date:  27.05.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.types import OneOrMany
from ampel.model.UnitModel import UnitModel

class T1CombineCompute(UnitModel[str]):
	"""
	Combine datapoints and request computation for the created t1 document
	"""

	#: Filter result codes that should trigger this T1.
	#: If not specified, T1 documents will be created in response to any passing alert.
	group: None | OneOrMany[int] = None

	#: T1 compute unit
	compute: UnitModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/util/template.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.07.2021
# Last Modified Date: 30.12.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from importlib import import_module
from functools import cache
from typing import Any
from collections.abc import Sequence
from ampel.types import JDict
from ampel.abstract.AbsTiedT2Unit import AbsTiedT2Unit
from ampel.model.UnitModel import UnitModel
from ampel.model.ingest.T2Compute import T2Compute
from ampel.config.builder.FirstPassConfig import FirstPassConfig


def check_tied_units(
	all_t2_units: list[T2Compute],
	first_pass_config: FirstPassConfig | JDict
) -> None:
	"""
	:raises: ValueError if tied t2 units are present in t2_units but the requred t2 units are not present in t2_compute
	"""
	tied_units: list[T2Compute] = []
	for el in all_t2_units:
		if "AbsTiedT2Unit" in first_pass_config['unit'][el.unit]['base']:
			tied_units.append(el)
		
	def as_unitmodel(t2_unit_model: T2Compute) -> UnitModel:
		return UnitModel(**{k: v for k, v in t2_unit_model.dict().items() if k in UnitModel.get_model_keys()})
		
	@cache
	def get_default_dependencies(unit: str) -> list[JDict]:
		klass: type[AbsTiedT2Unit] = getattr(
			import_module(
				first_pass_config['unit'][unit]['fqn']
			),
			unit
		)
		return [dep.dict() for dep in klass.t2_dependency]

	for tied_unit in tied_units:
		t2_deps = (
			(tied_unit.config if isinstance(tied_unit.config, dict) else {}) | \
			(tied_unit.override or {})
		).get("t2_dependency") or get_default_dependencies(tied_unit.unit)
		for t2_dep in t2_deps:
			dependency_config: UnitModel[str] = UnitModel(
				unit=t2_dep["unit"],
				config=t2_dep.get("config"),
				override=t2_dep.get("override")
			)
			candidates = [as_unitmodel(unit) for unit in all_t2_units if unit.unit == dependency_config.unit]
			if not any((c.dict() == dependency_config.dict() for c in candidates)):
				raise ValueError(
					f"Unit {tied_unit.unit} depends on unit {dependency_config.dict()}, "
					f"which was not configured. Possible matches are: "
					f"{[c.dict() for c in candidates]}"
				)

def filter_units(
	units: Sequence[UnitModel],
	abs_unit: str | list[str],
	config: FirstPassConfig | JDict
) -> list[JDict]:
	"""
	:returns: unit defintions (dict) that are subclass of the provided abstract class name.
	"""

	if isinstance(abs_unit, str):
		abs_unit = [abs_unit]

	return [
		el.dict(exclude_unset=True)
		for el in units
		if any(
			unit in config['unit'][el.unit].get("base", [])
			for unit in abs_unit
		)
	]


def resolve_shortcut(unit: str | JDict) -> JDict:
	return unit if isinstance(unit, dict) else {'unit': unit}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/ConfigBuilder.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.09.2019
# Last Modified Date:  24.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import os, sys, re, json, yaml, datetime, getpass, importlib, subprocess
from multiprocessing import Pool
from typing import Any
from collections.abc import Iterable

from ampel.log.utils import log_exception
from ampel.abstract.AbsChannelTemplate import AbsChannelTemplate
from ampel.log.AmpelLogger import AmpelLogger, VERBOSE, DEBUG, ERROR
from ampel.config.builder.DisplayOptions import DisplayOptions
from ampel.config.builder.FirstPassConfig import FirstPassConfig
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.config.collector.T02ConfigCollector import T02ConfigCollector
from ampel.template.ChannelWithProcsTemplate import ChannelWithProcsTemplate
from ampel.config.collector.ProcessConfigCollector import ProcessConfigCollector
from ampel.config.collector.ChannelConfigCollector import ChannelConfigCollector
from ampel.config.builder.ProcessMorpher import ProcessMorpher
from ampel.secret.AESecret import AESecret
from ampel.secret.AESecretProvider import AESecretProvider
from ampel.util.recursion import walk_and_process_dict
from ampel.util.mappings import get_by_path, set_by_path, dictify


class ConfigBuilder:
	"""
	Builds a central configuration dict for ampel. Config building is a two pass process:
	First, all available configuration files are loaded from different repositories
	and information merged together (into the instance self.first_pass_config)
	Then, the config is 'morphed' into its final structure by a multi-step process (method build_config)
	"""

	_default_processes = ["DefaultT2Process", "DefaultPurge"]

	def __init__(self, options: DisplayOptions, logger: None | AmpelLogger = None) -> None:

		self.logger = AmpelLogger.get_logger(
			console={'level': DEBUG if options.verbose else ERROR}
		) if logger is None else logger

		self.first_pass_config = FirstPassConfig(options, self.logger)
		self.templates: dict[str, Any] = {}
		self.verbose = options.verbose
		self.options = options
		self.error = False


	def load_ampel_conf(self,
		d: dict,
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> None:

		if self.verbose:
			self.logger.log(VERBOSE, f"Loading global ampel conf ({register_file}) from repo {dist_name}")

		# "mongo" "logging" "channel" "unit" "process" "alias" "resource"
		for k in self.first_pass_config.conf_keys:

			if k not in d:
				continue

			if k in ('unit', 'process', 'alias'):
				if isinstance(d[k], list):
					self.first_pass_config[k].add(d[k], dist_name, version, register_file)
				elif isinstance(d[k], dict):
					# kk = 't0', 't1', 't2', 't3', ... for root key "process" or "alias"
					for kk, v in d[k].items():
						if kk in self.first_pass_config[k]:
							if self.verbose:
								self.logger.log(VERBOSE, f"Parsing {k}.{kk}")
							self.first_pass_config[k][kk].add(v, dist_name, version, register_file)
						else:
							self.logger.error(f"Unknown config element: {k}.{kk}")

			else:
				self.first_pass_config[k].add(d[k], dist_name, version, register_file)

		if 'template' in d:
			self.register_channel_templates(d['template'], dist_name, version, register_file)


	def register_channel_templates(self,
		chan_templates: dict[str, str],
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> None:

		if not isinstance(chan_templates, dict):
			raise ValueError('Provided argument must be a dict instance')

		for k, v in chan_templates.items():

			if k in self.templates:
				raise ValueError('Duplicated channel template: ' + k)

			if self.verbose:
				self.logger.log(VERBOSE,
					f'Registering template "{k}" ' +
					register_file if register_file else '' +
					ConfigCollector.distrib_hint(distrib=dist_name)
				)

			self.templates[k] = getattr(
				importlib.import_module(v),
				v.split('.')[-1]
			)


	def build_config(self,
		stop_on_errors: int = 2,
		config_validator: None | str = "ConfigChecker",
		skip_default_processes: bool = False,
		json_serializable: bool = True,
		pwds: None | Iterable[str] = None,
		ext_resource: None | str = None,
		get_unit_env: bool = True,
		save: bool | str | None = None,
		sign: int = 6,
	) -> dict[str, Any]:
		"""
		Pass 2.
		Builds the final ampel config using previously collected config pieces (contained in self.first_pass_config)
		This involves a multi-step process where the config is 'morphed' into its final structure.

		:param stop_on_errors: by default, config building stops and raises an exception if an error occured.
			- 2: stop on errors
			- 1: ignore errors in first_pass_config only (will stop on morphing/scoping/template errors)
			- 0: ignore all errors
		Note that issues might/will later arise with your ampel system.

		:param get_unit_env: whether to get a full list of dependencies for each ampel unit (for trace ids)

		:param pwds: config section 'resource' might contain AES encrypted entries.
		If passwords are provided to this method, thoses entries will be decrypted.

		:param ext_resource: path to resource config file (yaml) to be integrated into the final ampel config.
		The file must not contain the key "resource:", only content (ex: "mongo: mongodb://localhost:27050")
		Note that resources defined externally (by this config file) will prevale over resources (with the same keys)
		gathered from the repositories. "Secrets" can be used, they will be resolved during config building
		if AES keys are provided with the parameter 'pwds', otherwise, they will be decrypted at run time just as with
		regular resource.

		:param json_serializable: if True, stringify int keys in section 'confid' and potential int channel names.

		:param skip_default_processes: set to True to discard default processes defined by ampel-core.
		The static variable ConfigBuilder._default_processes references those processes by name.
		Set skip_default_processes=True if your repositories define their own default T2/T3 processes.

		:param sign: append truncated file signature (last n digits) to filename. Ex: ampel_conf_4a72fd.yaml

		:raises: ValueError if self.error is True - this behavior can be disabled using the parameter stop_on_errors
		"""

		if self.first_pass_config.has_nested_error():
			if stop_on_errors > 1:
				raise ValueError(
					'Error were reported in first pass config, you can use the option stop_on_errors = 1 (or 0)\n' +
					'to bypass this exception and get the (possibly non-working) config nonetheless'
				)

		if ext_resource and not os.path.exists(ext_resource):
			raise ValueError(f"External resource file not found: '{ext_resource}'")

		out = {
			k: self.first_pass_config[k]
			for k in FirstPassConfig.conf_keys.keys()
		}

		out['process'] = {}

		if self.verbose:
			self.logger.log(VERBOSE, 'Getting unit dependencies')

		if get_unit_env:

			all_deps: dict[str, str | float] = {}
			env = os.environ.get('CONDA_DEFAULT_ENV')
			out['environment'] = {f'conda_{env}' if env else 'default': all_deps}

			with Pool() as pool:
				for res in pool.starmap(
					get_unit_dependencies,
					[(el['fqn'], env) for el in out['unit'].values()]
				):
					if self.verbose:
						self.logger.log(VERBOSE, f'{res[0]} dependencies: {res[1] or None}')
					if res[1]:
						out['unit'][res[0]]['dependencies'] = list(res[1].keys())
						all_deps.update(res[1])

		# Add t2 init config collector (in which both hashed values of t2 run configs
		# and t2 init config will be added)
		out['confid'] = T02ConfigCollector('confid', self.options, logger=self.logger)

		# Add (possibly transformed) processes to output config
		for tier in (0, 1, 2, 3, "ops"):

			tier_name = tier if tier == "ops" else f"t{tier}"

			if self.verbose:
				self.logger.log(VERBOSE, f'Checking standalone {tier_name} processes')

			p_collector = ProcessConfigCollector(
				'process', self.options, tier=tier, logger=self.logger # type: ignore[arg-type]
			)

			#out['process'][f't{tier}'] = {
			#	k: self.first_pass_config[f't{tier}'][k]
			#	for k in FirstPassConfig.tier_keys if k != 'process'
			#}

			# New empty process collector to gather morphed processes
			out['process'][tier_name] = p_collector

			# For each process collected before, apply transformations
			# and add it to our (almost) final process collector.
			# 'almost' because a gathering of T0 processes may occure later
			for p in self.first_pass_config['process'][tier_name].values():

				if skip_default_processes and p["name"] in self._default_processes:
					self.logger.info(f'Skipping {tier_name} default processes: {p["name"]}')
					continue

				if self.verbose:
					self.logger.log(VERBOSE, f'Morphing standalone {tier_name} processes: {p["name"]}')
					pass

				try:
					p_collector.add(
						self.new_morpher(p) \
							.scope_aliases(self.first_pass_config) \
							.apply_template(self.first_pass_config) \
							.hash_t2_config(out) \
							.generate_version(self.first_pass_config) \
							.get(),
						p.get('distrib'),
						p.get('version'),
						p.get('source')
					)
				except Exception as e:
					self.logger.error(f'Unable to morph process {p["name"]}', exc_info=e)
					if stop_on_errors > 0:
						raise e

		# Setup empty channel collector
		out['channel'] = ChannelConfigCollector('channel', self.options, logger=self.logger)
		morph_errors = []

		# Add (possibly transformed) channels
		for chan_name, chan_dict in self.first_pass_config['channel'].items():

			if not chan_dict.get('active', False):
				self.logger.info(f'Ignoring deactivated channel: {chan_name}')
				continue

			tpl = None

			# Template processing is required for this particular channel
			try:
				tpl = self._get_channel_tpl(chan_dict)
			except Exception as ee:
				log_exception(self.logger, msg=f'Unable to load template ({chan_name})', exc=ee)
				if stop_on_errors > 0:
					raise ee
				continue

			# Template processing is required for this particular channel
			if tpl:

				# Extract channel definition from template instance
				try:
					out['channel'].add(
						(c := tpl.get_channel(self.logger)),
						c.get('distrib'), c.get('version'), c.get('source')
					)
				except Exception as ee:
					log_exception(self.logger, msg=f'Unable to get channel from template ({chan_name})', exc=ee)
					if stop_on_errors > 0:
						raise ee
					continue

				# Extract process definition from template instance
				ps = []
				try:
					ps = tpl.get_processes(self.logger, self.first_pass_config)
				except Exception as ee:
					log_exception(self.logger, msg=f'Unable to get processes from template ({chan_name})', exc=ee)
					if stop_on_errors > 0:
						raise ee
					continue

				# Retrieve processes possibly embedded in channel def
				for p in ps:

					if self.verbose:
						self.logger.log(VERBOSE,
							f'Morphing channel embedded t{p["tier"]} process: {p["name"]}'
						)

					try:
						# Add transformed process to final process collector
						out['process'][f't{p["tier"]}'].add(
							self.new_morpher(p) \
								.scope_aliases(self.first_pass_config) \
								.apply_template(self.first_pass_config) \
								.hash_t2_config(out) \
								.enforce_t3_channel_selection(chan_name) \
								.generate_version(self.first_pass_config) \
								.get(),
							p.get('distrib'),
							p.get('version'),
							p.get('source')
						)
					except Exception as ee:
						morph_errors.append(p["name"])
						log_exception(
							self.logger, exc=ee,
							msg=f'Unable to morph embedded process {p["name"]} (from {p["source"]})'
						)
						if stop_on_errors > 0:
							raise ee

			else:

				# Raw/Simple/Standard channel definition
				# (encouraged behavior actually)
				out['channel'].add(
					chan_dict, chan_dict.get('distrib'),
					chan_dict.get('version'), chan_dict.get('source')
				)

		if ext_resource:
			with open(ext_resource, "r") as f:
				out['resource'].update(
					yaml.safe_load(f)
				)

		# Optionaly decrypt aes encrypted config entries
		if pwds:

			self.logger.info('Resolving AES secrets')
			sp = AESecretProvider(pwds)
			enc_confs: list[tuple[dict, str, AESecret, str]] = []
			walk_and_process_dict(
				arg = out,
				callback = self._gather_aes_config_callback,
				enc_confs = enc_confs
			)

			for tup in enc_confs:
				self.logger.info(f"Resolving {tup[3]}")
				d = tup[0]
				k = tup[1]
				secret = tup[2]
				if not sp.tell(secret, str):
					self.logger.info(" -> Secret not resolvable with specified password(s)")
				else:
					d[k] = secret.get()

	
		# Register templates in config (might be used by the 'ampel job' CLI)
		out['template'] = {k: v.__module__ for k, v in self.templates.items()}
		self.logger.info('Done building config')

		# Error Summary
		if out['unit'].err_fqns:
			self.logger.break_aggregation()
			self.logger.info('Erroneous units (import failed):')
			for el in out['unit'].err_fqns:
				self.logger.info(f"{el[0]} - [{el[1].__class__.__name__}] {el[1]}")

		if morph_errors:
			self.logger.break_aggregation()
			self.logger.info('Erroneous process definitions (morphing failed):')
			for el in morph_errors:
				self.logger.info(el) # type: ignore[arg-type]
		
		# Cast into plain old dicts
		d = {
			'build': {
				'date': (now := datetime.datetime.now()).strftime("%d/%m/%Y"),
				'time': now.strftime("%H:%M:%S"),
				'by': getpass.getuser()
			}
		} | dictify(out)

		# Cosmetic: sort units first by category, then alphabetically
		u = d['unit']
		d['unit'] = {}
		for el in ("AbsProcessController", "AbsEventUnit", "ContextUnit", "LogicalUnit"):
			dd = {k: u[k] for k in sorted(u.keys()) if el in u[k]['base']}
			d['unit'] |= dd
			for k in dd:
				del u[k]
		d['unit'] |= u # Aux units

		# Convert int keys to str (ensures JSON compatibility)
		if json_serializable:

			for k in [el for el in out['channel'].keys() if isinstance(el, str) and el.isdigit()]:
				out['channel'][str(k)] = out['channel'].pop(k)
			for k in list(out['confid'].keys()):
				out['confid'][str(k)] = out['confid'].pop(k)

		if config_validator:

			from importlib import import_module
			validator = getattr(
				import_module("ampel.config.builder." + config_validator),
				config_validator
			)(d, self.logger, self.verbose)
			return validator.validate()

		if save:

			import pathlib # type: ignore
			path = pathlib.Path(save if isinstance(save, str) else 'ampel_conf.yaml')
			with open(path, 'w') as file:
				yaml.dump(d, file, sort_keys=False)

			if sign:
				import hashlib
				h = hashlib.blake2b(path.read_bytes()).hexdigest()[:sign]
				path = path.rename(path.with_stem(f"{path.stem}_{h}"))

			self.logger.break_aggregation()
			with open(path, "r") as file:
				self.logger.info(f'Config file saved as {path} [{len(file.readlines())} lines]')

		return d


	def new_morpher(self, process: dict[str, Any]) -> ProcessMorpher:
		"""
		Returns an instance of ProcessMorpher using the provided
		process dict and the internal logger and templates
		"""
		return ProcessMorpher(
			process, self.templates, self.logger, self.verbose
		)


	def _get_channel_tpl(self, chan_dict: dict[str, Any]) -> None | AbsChannelTemplate:
		"""
		Internal method used to check if a template (and which one)
		should be applied to a given channel dict.
		"""

		if 'template' in chan_dict:

			if self.verbose:
				self.logger.log(VERBOSE,
					f'Channel {chan_dict["channel"]} ({chan_dict["source"]}) '
					f'declares template: {chan_dict["template"]}'
				)

			if chan_dict['template'] not in self.templates:
				raise ValueError(f'Unknown template: {chan_dict["template"]}')

			return self.templates[chan_dict['template']](**chan_dict)


		if 'process' in chan_dict:
			return ChannelWithProcsTemplate(**chan_dict)

		return None


	def gather_processes(self,
		config: FirstPassConfig, tier: int, match: str, collect: str
	) -> None | dict:
		"""
		:param channel_names:
		- None: all the available channels from the ampel config will be loaded
		- String: channel with the provided id will be loaded
		- List of strings: channels with the provided ids will be loaded
		"""

		processes: list[dict] = [
			el for el in config[f't{tier}']['process'].values()
			if re.match(match, el.get('name'))
		]

		if len(processes) <= 1:
			return None

		init_configs = []
		dist_names: set[str] = set()
		out_proc: None | dict = None

		for p in processes:

			if not p.get('active', False):
				self.logger.info(f'Ignoring deactivated process: {p["name"]}')
				continue

			# Use first active proces as template for the multi-channel process
			# to be generated by this method
			if not out_proc:
				# Faster deep copy
				out_proc = json.loads(json.dumps(p))

			# Inactivate this particular process as it will be run in a multi-channel process
			self.logger.info(f'Deactivating process: {p["name"]}')
			p['active'] = False

			# Gather init config
			init_configs.append(
				get_by_path(p, collect)
			)

			# Collect distribution name
			if p.get('distrib'):
				dist_names.add(p.get('distrib')) # type: ignore

		if out_proc is None:
			return None

		# for T0 processes: collect=processor.config.directives
		set_by_path(out_proc, collect, init_configs)
		out_proc['name'] = match
		out_proc['distrib'] = '/'.join(dist_names)

		return out_proc


	def print(self) -> None:
		self.first_pass_config.print()


	def _gather_aes_config_callback(self, path, k, d, **kwargs) -> None:

		if d[k] and 'iv' in d[k]:
			try:
				secret = AESecret(**k[d])
			except Exception:
				return
			# dict, key, secret, string path (debug)
			kwargs['enc_confs'].append((d, k, secret, f"{path}.{k}"))


def get_unit_dependencies(fqn: str, env: None | str) -> tuple[str, dict]:

	if env:
		ret = subprocess.run(
			f"""eval "$(conda shell.bash hook)"
				conda activate {env}
				python3 -m 'ampel.config.builder.get_env' {fqn}
			""",
			stdout = subprocess.PIPE,
			shell = True,
			check = True
		)
	else:
		ret = subprocess.run(
			[sys.executable, '-m', 'ampel.config.builder.get_env', fqn],
			stdout = subprocess.PIPE,
			check = True
		)

	return fqn.split(".")[-1], eval(ret.stdout.decode("utf-8"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsEventUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                04.02.2020
# Last Modified Date:  25.07.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.types import ChannelId, OneOrMany, Traceless
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod, defaultmethod
from ampel.core.EventHandler import EventHandler
from ampel.core.ContextUnit import ContextUnit
from ampel.enum.EventCode import EventCode
from ampel.log.LogFlag import LogFlag


class AbsEventUnit(AmpelABC, ContextUnit, abstract=True):
	"""
	Base class for units orchestrating an ampel event.
	Ampel processes (from the ampel config) are the most common events.
	They define specifications of an event using a standardized model/schema:
	which subclass of this abstract class is to be run by a given controller
	at (a) given time(s) using a specific configuration.

	An ampel event must be registered and traceable and thus generates:
	- a new run-id
	- a new event document (in the event collection)
	- possibly Log entries (less is better)
	- new trace ids if not registered previously

	Subclassses of this class typically modify tier documents
	whereby the changes originate from results of logical unit(s)
	instantiated by this class.
	"""

	#: name of the associated process
	process_name: Traceless[str]

	#: hash of potentially underlying job schema
	job_sig: None | int = None

	#: channels associated with the process
	channel: None | OneOrMany[ChannelId] = None

	#: raise exceptions instead of catching and logging
	#: if True, troubles collection will not be populated if an exception occurs
	raise_exc: bool = False

	#: logging configuration to use, from the logging section of the Ampel config
	log_profile: str = "default"

	#: flag to apply to all log records created
	base_log_flag: LogFlag = LogFlag.SCHEDULED_RUN

	#: optional additional kwargs to pass to :class:`~ampel.mongo.update.var.DBLoggingHandler.DBLoggingHandler`
	db_handler_kwargs: None | dict[str, Any] = None

	#: Some subclasses allow for non-serializable input parameter out of convenience (jupyter notebooks).
	#: For example, an AlertSupplier instance can be passed as argument of an AlertConsumer.
	#: Unless a custom serialization is implemented, this disables the hashing of input parameters
	#: and thus the generation of "trace ids".
	#: Set provenance to False to ignore trace ids generation error which will be raised otherwise.
	provenance: bool = True


	@defaultmethod
	def prepare(self, event_hdlr: EventHandler) -> None | EventCode:
		"""
		Gives units the opportunity to cancel the subsequent call to the proceed() method
		by returning EventCode.PRE_CHECK_EXIT (which avoids 'burning' a run id)
		"""
		return None


	@abstractmethod(var_args=True)
	def proceed(self, event_hdlr: EventHandler) -> Any:
		...


	def run(self, event_hdlr: None | EventHandler = None) -> Any:

		if event_hdlr is None:
			event_hdlr = EventHandler(
				self.process_name,
				self.context.get_database(),
				job_sig = self.job_sig,
				raise_exc = self.raise_exc
			)

		# Give units opportunity to cancel the run (T2 worker when no tickets are avail for example)
		pre_check = self.prepare(event_hdlr)
		if pre_check == EventCode.PRE_CHECK_EXIT:
			event_hdlr.set_code(pre_check)
			event_hdlr.register()
			return None

		event_hdlr.register(
			run_id = self.context.new_run_id()
		)

		ret = None
		try:
			if ret := self.proceed(event_hdlr):
				event_hdlr.add_extra(ret=ret)
		except Exception as e:
			if self.raise_exc:
				raise e
			event_hdlr.code == EventCode.EXCEPTION

		# Set default event code if sub-class didn't customize it
		if event_hdlr.code is None:
			event_hdlr.code = EventCode.OK

		# Update duration and code in event doc
		event_hdlr.update()

		# Forward returned run() value to caller
		return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/ingest/T2Compute.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                10.03.2020
# Last Modified Date:  28.09.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Generic
from ampel.types import OneOrMany
from ampel.model.UnitModel import UnitModel, T
from ampel.model.DPSelection import DPSelection

class T2Compute(UnitModel[T], Generic[T]):

	#: Ingester options
	ingest: None | str | DPSelection = None

	#: Filter result codes that should trigger this T2. If not specified, T2
	#: documents will be created in response to any passing alert.
	group: OneOrMany[int] = []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t1/T1SimpleCombiner.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                01.01.2018
# Last Modified Date:  14.06.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Iterable
from ampel.content.DataPoint import DataPoint
from ampel.types import DataPointId
from ampel.abstract.AbsT1CombineUnit import AbsT1CombineUnit


class T1SimpleCombiner(AbsT1CombineUnit):

	def combine(self, datapoints: Iterable[DataPoint]) -> list[DataPointId]:
		"""
		:param datapoints: dict instances representing datapoints
		"""

		if self.channel:
			return [
				dp['id'] for dp in datapoints
				if not("excl" in dp and self.channel in dp['excl'])
			]

		return [dp['id'] for dp in datapoints]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/T3AdaptativeStager.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                06.01.2020
# Last Modified Date:  09.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from itertools import islice
from multiprocessing import JoinableQueue
from multiprocessing.pool import ThreadPool, AsyncResult
from collections.abc import Generator, Sequence

from ampel.types import ChannelId
from ampel.view.T3Store import T3Store
from ampel.view.SnapView import SnapView
from ampel.model.UnitModel import UnitModel
from ampel.content.T3Document import T3Document
from ampel.log import VERBOSE
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.abstract.AbsT3Filter import AbsT3Filter
from ampel.abstract.AbsT3Projector import AbsT3Projector
from ampel.t3.stage.T3ThreadedStager import T3ThreadedStager
from ampel.t3.stage.T3ProjectingStager import RunBlock
from ampel.t3.stage.ThreadedViewGenerator import ThreadedViewGenerator


class T3AdaptativeStager(T3ThreadedStager):
	"""
	Unit stager that for each channel found in the elements loaded by the previous stages:
	
	- spawns a dedicated :class:`~ampel.t3.run.T3ProjectingStager.T3ProjectingStager` instance configured to filter and project elements wrt this channel
	- execute the associated T3 units

	Example:
	A general T3 process performs a broad, channel-less query.
	Many stocks are returned, each possibly associated with different channels.
	This unit builds a set of all channels referenced by the results.
	Then, for each channel, the information about the referenced elements are posted into
	a dedicated Slack channel whose name corresponds to the Ampel channel.
	Suppose, for example, that the results contain objects:

	- ``<object 1 with channel {A}>``
	- ``<object 2 with channel {A, B}>``
	- ``<object 3 with channel {B, C, D}>``

	This will cause:
	
	- Object 1 and 2 to be posted to slack channel #A
	- Object 2 and 3  to be posted to slack channel #B
	- Object 3 to be posted to slack channel #C
	- Object 3 to be posted to slack channel #D
	"""

	execute: Sequence[UnitModel]
	white_list: None | set[str] = None
	black_list: None | set[str] = None

	remove_empty: bool = True
	unalterable: bool = True
	freeze: bool = False

	nthread: int = 4


	def __init__(self, **kwargs) -> None:
		"""
		:param execute: sequence of T3 units to be executed
		:param white_list: perform operations only for channels referenced in this list
		:param black_list: do nothing for channels referenced in this list
		:param channel: (from super class T3UnitRunner): if set, all log entries produced
		by the underlying t3 units will be associated with the provided channel.
		Otherwise, log entries will be associated with the projected channel.
		:param remove_empty: see T3BaseProjector docstring
		:param unalterable: see T3ChannelProjector docstring
		:param freeze: see T3ChannelProjector docstring
		"""

		super().__init__(**kwargs)
		self.run_blocks: dict[ChannelId, RunBlock] = {}

		if self.logger.verbose > 1 and (self.white_list or self.black_list):
			self.warned: set[ChannelId] = set()

		if self.black_list and self.white_list:
			raise ValueError("Can't have both black and white lists")

		self.queues: dict[AbsT3ReviewUnit, JoinableQueue[SnapView]] = {}
		self.generators: list[ThreadedViewGenerator] = []
		self.async_results: list[AsyncResult] = []


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> None | Generator[T3Document, None, None]:

		ts = time()
		with ThreadPool(processes=self.nthread) as pool:

			# Chunk input buffers (loaded from generator)
			while (data := list(islice(gen, self.chunk_size))):

				# step 1: Gather all channels
				# (will raise Error if stock documents were to requested to be loaded)
				channels = self.filter_channels({
					ell for el in data for ell in el['stock']['channel'] # type: ignore[index]
				})

				# step 2: spawn T3UnitRunner instances
				for chan in channels:

					if chan in self.run_blocks:
						rb = self.run_blocks[chan]

					else:

						rb = RunBlock()
						rb.units = [self.get_unit(um, chan=chan) for um in self.execute]

						# Create and start T3 units "process(...)" threads (generator will block)
						qs, gs, rs = self.create_threaded_generators(pool, rb.units, t3s)

						self.queues.update(qs)
						self.generators.extend(gs)
						self.async_results.extend(rs)

						rb.filter = self.context.loader.new(
							model = UnitModel(unit="T3AmpelBufferFilter", config={"channel": chan}),
							unit_type = AbsT3Filter,
							logger = self.logger
						)

						if self.save_stock_ids:
							rb.stock_ids = []

						rb.projector = AuxUnitRegister.new_unit(
							model = UnitModel(
								unit="T3ChannelProjector",
								config={"channel": chan, "remove_empty": self.remove_empty}
							),
							sub_type = AbsT3Projector,
							logger = self.logger
						)

						# Dict used to potentially optimize views generation
						rb.qdict = {}
						for unit in rb.units:
							if unit.__class__._View not in rb.qdict:
								rb.qdict[unit.__class__._View] = []
							rb.qdict[unit.__class__._View].append(qs[unit])

						self.run_blocks[chan] = rb

					buffers: Sequence[AmpelBuffer] = data

					if self.logger.verbose:
						self.logger.log(VERBOSE, "Applying run-block filter")

					buffers = rb.filter.filter(buffers) # type: ignore[union-attr]

					if self.save_stock_ids:
						rb.stock_ids.extend([el['id'] for el in buffers])  # type: ignore[union-attr]

					if self.logger.verbose:
						self.logger.log(VERBOSE, "Applying run-block projection")

					buffers = rb.projector.project(buffers) # type: ignore[union-attr]

					self.put_views(buffers, rb.qdict)

			# Send sentinel all threaded generators
			for q in self.queues.values():
				q.put(None) # type: ignore[arg-type]

			for async_res, generator, t3_unit in zip(self.async_results, self.generators, list(self.queues.keys())):

				# potential T3Record to be included in the T3Document
				if (t3_unit_result := async_res.get()):
					if (z := self.handle_t3_result(t3_unit, t3_unit_result, t3s, generator.stocks, ts)):
						yield z

				self.flush(t3_unit)


	def filter_channels(self, channels: set[ChannelId]) -> set[ChannelId]:

		if self.white_list:
			if self.logger.verbose > 1:
				tmp = channels & self.white_list
				for chan in channels - self.white_list:
					if chan not in self.warned:
						self.logger.debug(f"Ignoring channel {chan} (not in white list)")
						self.warned.add(chan)
				channels = tmp
			else:
				channels = channels & self.white_list

		if self.black_list:
			if self.logger.verbose > 1:
				tmp = channels - self.black_list
				for chan in channels & self.black_list:
					if chan not in self.warned:
						self.logger.debug(f"Ignoring channel {chan} (in black list)")
						self.warned.add(chan)
				channels = tmp
			else:
				channels = channels - self.black_list

		return channels



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/UnitConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  23.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import os, sys, re, importlib, traceback
from typing import Any
from os.path import sep
from contextlib import contextmanager

from ampel.protocol.LoggingHandlerProtocol import AggregatingLoggingHandlerProtocol
from ampel.log.handlers.AmpelStreamHandler import AmpelStreamHandler
from ampel.util.collections import ampel_iter
from ampel.util.distrib import get_files
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.log import VERBOSE


class RemoteUnitDefinition(AmpelBaseModel):
	class_name: str
	base: list[str]


class UnitConfigCollector(ConfigCollector):


	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self.err_fqns: list[tuple[str, Exception]] = []


	def add(self,
		arg: list[dict[str, str] | str],
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> None:

		# Cosmetic
		if isinstance(self.logger.handlers[0], (AggregatingLoggingHandlerProtocol, AmpelStreamHandler)):
			agg_int = self.logger.handlers[0].aggregate_interval
			self.logger.handlers[0].aggregate_interval = 1000

		# tolerate list containing only 1 element defined as dict
		for el in ampel_iter(arg):

			try:

				if isinstance(el, str):

					# Package definition (ex: ampel.ztf.t3)
					# Auto-load units in defined package
					if el.split('.')[-1][0].islower():
						package_path = el.replace(".", sep)
						try:
							for fpath in get_files(dist_name, lookup_dir=package_path):
								if sep in fpath.replace(package_path + sep, ""):
									continue
								self.add(
									[fpath.replace(sep, ".").replace(".py", "")],
									dist_name, version, register_file
								)
						except Exception:
							self.logger.info(f'Units auto-registration has failed for package {el}')
						continue

					# Standart unit definition (ex: ampel.t3.stage.T3AggregatingStager)
					else:
						class_name = self.get_class_name(el)
						if not (mro := self.get_mro(el, class_name)):
							self.logger.break_aggregation()
							continue
						entry: dict[str, Any] = {
							'fqn': el,
							'base': mro,
							'distrib': dist_name,
							'file': register_file,
							'version': version
						}

				elif isinstance(el, dict):
					try:
						d = RemoteUnitDefinition(**el)
					except Exception:
						self.error(
							'Unsupported unit definition (dict)' +
							self.distrib_hint(dist_name, register_file)
						)
						continue

					class_name = d.class_name
					entry = {
						'base': d.base,
						'distrib': dist_name,
						'file': register_file,
						'version': version
					}

				else:
					self.error(
						'Unsupported unit config format' +
						self.distrib_hint(dist_name, register_file)
					)
					continue

				if self.get(class_name):
					self.duplicated_entry(
						conf_key = class_name,
						section_detail = f'{self.tier} {self.conf_section}',
						new_file = register_file,
						new_dist = dist_name,
						prev_file = self.get(class_name).get('conf', 'unknown') # type: ignore
					)
					continue

				if "AmpelUnit" not in entry['base'] and "AmpelBaseModel" not in entry['base']:
					self.logger.info(
						f'Unrecognized base class for {self.conf_section} {class_name} ' +
						self.distrib_hint(dist_name, register_file)
					)
					continue

				for base in ("AmpelABC", "AmpelUnit"):
					if base in entry["base"]:
						entry["base"].remove(base)

				if self.options.verbose:
					self.logger.log(VERBOSE,
						f'Adding {self.conf_section}: {class_name}'
					)

				self.__setitem__(class_name, entry)

			except Exception as e:

				if 'class_name' not in locals():
					class_name = 'Unknown'

				self.error(
					f'Error occured while loading {self.conf_section} {class_name} ' +
					self.distrib_hint(dist_name, register_file),
					exc_info=e
				)

		if isinstance(self.logger.handlers[0], AggregatingLoggingHandlerProtocol):
			self.logger.handlers[0].aggregate_interval = agg_int


	@staticmethod
	def get_class_name(fqn: str) -> str:
		"""
		Returns class name for a given fully qualified name
		Note: we here assume the ampel convention that module name equals class name
		(i.e that we have one class per module)
		"""
		return re.sub(r'.*\.', '', fqn) # noqa


	def get_mro(self, module_fqn: str, class_name: str) -> None | list[str]:
		"""
		:param module_fqn: fully qualified name of module
		:param class_name: declared class name in the module specified by "module_fqn"
		:returns: the method resolution order (except for the first and last members
		that is of no use for our purpose) of the specified class
		"""

		try:
			# contextlib.redirect_stderr does not work with C-loaded backends (ex: annoying healpix warnings)
			with stderr_redirected(self.options.hide_stderr):
				UnitClass = getattr(
					# import using fully qualified name
					importlib.import_module(module_fqn),
					class_name
				)
		except Exception as e:

			self.err_fqns.append((module_fqn, e))

			if self.options.hide_module_not_found_errors and isinstance(e, ModuleNotFoundError):
				return None

			# Suppress superfluous traceback entries
			print("", file=sys.stderr)
			print("Cannot import " + module_fqn, file=sys.stderr)
			print("-" * (len(module_fqn) + 14), file=sys.stderr)
			for el in traceback.format_exc().splitlines():
				if "_bootstrap" not in el and "in import_module" not in el:
					print(el, file=sys.stderr)

			return None

		return [
			el.__name__
			for el in UnitClass.__mro__[:-1]
			if 'ampel' in el.__module__
		]


# contextlib.redirect_stderr does not work with C-loaded backends (ex: annoying healpix warnings)
# Code below is borrowed from:
# https://stackoverflow.com/questions/4675728/redirect-stdout-to-a-file-in-python/22434262#22434262
def fileno(file_or_fd):
	fd = getattr(file_or_fd, 'fileno', lambda: file_or_fd)()
	if not isinstance(fd, int):
		raise ValueError("Expected a file (`.fileno()`) or a file descriptor")
	return fd


@contextmanager
def stderr_redirected(activated: bool = False):

	if not activated:
		yield
		return

	stderr = sys.stderr
	stderr_fd = fileno(stderr)
	# copy stderr_fd before it is overwritten
	# NOTE: `copied` is inheritable on Windows when duplicating a standard stream
	with os.fdopen(os.dup(stderr_fd), 'wb') as copied:
		stderr.flush()  # flush library buffers that dup2 knows nothing about
		try:
			os.dup2(fileno(os.devnull), stderr_fd)  # $ exec >&os.devnull
		except ValueError:  # filename
			with open(os.devnull, 'wb') as to_file:
				os.dup2(to_file.fileno(), stderr_fd)  # $ exec > os.devnull
		try:
			yield stderr # allow code to be run with the redirected stderr
		finally:
			# restore stderr to its previous value
			# NOTE: dup2 makes stderr_fd inheritable unconditionally
			stderr.flush()
			os.dup2(copied.fileno(), stderr_fd)  # $ exec >&copied



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/JobModel.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  15.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from functools import partial
from typing import Any, Literal
from ampel.model.ChannelModel import ChannelModel
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.job.ExpressionParser import ExpressionParser
from ampel.model.job.MongoOpts import MongoOpts
from ampel.model.job.TemplateUnitModel import TemplateUnitModel
from ampel.model.job.TaskUnitModel import TaskUnitModel
from ampel.model.job.EnvSpec import EnvSpec
from ampel.model.job.Parameter import Parameter
from ampel.model.job.utils import transform_expressions


class JobModel(AmpelBaseModel):

	name: None | str
	requirements: list[str] = []
	env: dict[str, EnvSpec] = {}
	channel: list[ChannelModel] = []
	alias: dict[Literal["t0", "t1", "t2", "t3"], dict[str, Any]] = {}
	parameters: list[Parameter] = []
	mongo: MongoOpts = MongoOpts()
	task: list[TemplateUnitModel | TaskUnitModel]


	def __init__(self, **kwargs):
		# Allow 'name' channel name alias to avoid 'channel[].channel'
		for c in kwargs.get('channel', []):
			if isinstance(c, dict) and 'name' in c:
				c["channel"] = c.pop("name")
		super().__init__(**kwargs)

	def resolve_expressions(
		self,
		target: dict,
		task: TaskUnitModel | TemplateUnitModel,
		item: None | str | dict | list = None
	) -> dict:
		"""
		Resolve any expressions of the form {{ expr }} found in string values of
		the target dict

		Supported expressions:

		- job parameters, e.g. {{ job.parameters.param }} resolves to the
		  job-level parameter "param"
		- task outputs, e.g. {{ task.step-0.outputs.parameters.token }} resolves
		  to the contents of the output file declared as "token" by the task
		  named "step-0"
		- iteration items, e.g. {{ item }} for a string-valued item or {{
		  item.name }} for dict-valued
		"""

		context = {
			"job": {
				"parameters": {param.name: param.value for param in self.parameters}
			},
			"task": {
				task.title: {
					"outputs": {
						"parameters": {
							spec.name: spec.value_from.path
							for spec in task.outputs.parameters
						}
					}
				}
				for task in self.task
			},
			"inputs": {
				"parameters": {
					param.name: param.value for param in task.inputs.parameters
				},
				"artifacts": {spec.name: spec.path for spec in task.inputs.artifacts},
			},
			"item": item,
		}

		return transform_expressions(
			target,
			transformation=partial(ExpressionParser.evaluate, context=context),
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/FirstPassConfig.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  18.06.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import json
from typing import Any, Literal
from ampel.log.AmpelLogger import AmpelLogger

from ampel.config.builder.DisplayOptions import DisplayOptions
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.config.collector.LoggingCollector import LoggingCollector
from ampel.config.collector.DBConfigCollector import DBConfigCollector
from ampel.config.collector.UnitConfigCollector import UnitConfigCollector
from ampel.config.collector.AliasConfigCollector import AliasConfigCollector
from ampel.config.collector.ProcessConfigCollector import ProcessConfigCollector
from ampel.config.collector.ChannelConfigCollector import ChannelConfigCollector
from ampel.config.collector.ResourceConfigCollector import ResourceConfigCollector
from ampel.config.collector.ForwardProcessConfigCollector import ForwardProcessConfigCollector

tiers: tuple[Literal[0, 1, 2, 3], ...] = (0, 1, 2, 3)

class FirstPassConfig(dict):
	"""
	Class used to aggregate config pieces into a central configuration dict for ampel.
	"""

	conf_keys: dict[str, None | type[ConfigCollector]] = {
		"mongo": DBConfigCollector,
		"logging": LoggingCollector,
		"channel": ChannelConfigCollector,
		"unit": None,
		"process": None,
		"alias": None,
		"resource": ResourceConfigCollector,
	}

	def __init__(self, options: DisplayOptions, logger: None | AmpelLogger = None) -> None:

		self.logger = AmpelLogger.get_logger() if logger is None else logger

		d: dict[str, Any] = {
			k: Klass(conf_section=k, options=options, logger=logger)
			for k, Klass in self.conf_keys.items() if Klass
		}

		d['pwd'] = []
		d['unit'] = UnitConfigCollector(
			conf_section="unit", options=options, logger=logger
		)

		# Allow process to be defined in root key
		d['process'] = ForwardProcessConfigCollector(
			root_config=self, conf_section="process", # type: ignore
			target_collector_type=ProcessConfigCollector,
			logger=logger, options=options
		)

		d['alias'] = {}
		for k in tiers:

			d['alias'][f"t{k}"] = AliasConfigCollector(
				conf_section='alias', options=options, logger=logger, tier=k
			)

			# Allow processes to be defined in sub-tier entries already (process.t0, process.t1, ...)
			d['process'][f"t{k}"] = ProcessConfigCollector(
				conf_section='process', options=options, logger=logger, tier=k
			)

		d['process']["ops"] = ProcessConfigCollector(
			conf_section='process', options=options, logger=logger, tier="ops"
		)

		super().__init__(d)


	def unset_errors(self, d: None | dict = None) -> None:
		""" """
		for v in d.values() if d is not None else self.values():
			if isinstance(v, dict):
				if getattr(v, 'has_error', False):
					v.has_error = False # type: ignore
				self.unset_errors(v)


	def has_nested_error(self, d=None, k=None) -> bool:

		ret = False

		for kk, dd in d.items() if d is not None else self.items():
			if isinstance(dd, dict):
				if getattr(dd, 'has_error', False):
					ret = True
					hint = f"(key: '{kk}')" if kk else ""
					self.logger.warn(f"{dd.__class__.__name__} {hint} has errors")
				if self.has_nested_error(dd, kk):
					ret = True

		return ret


	def print(self) -> None:

		if self.has_nested_error():
			self.logger.warn(
				"Warning: error were reported while collecting configurations"
			)

		self.logger.info(json.dumps(self, indent=4))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/view/MongoOneView.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.03.2021
# Last Modified Date:  07.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.types import ChannelId
from ampel.mongo.view.AbsMongoView import AbsMongoView

class MongoOneView(AbsMongoView):

	channel: ChannelId
	t0_has_chan: bool = True

	def stock(self) -> list[dict[str, Any]]:

		return [
			{'$match': {'channel': self.channel}},
			{
				'$addFields': {
					'ts': f"$ts.{self.channel}",
					'journal': self.filter_journal('journal')
				}
			},
			{'$project': {'channel': 0, 'journal.channel': 0}},
		]


	def t0(self) -> list[dict[str, Any]]:
		if self.t0_has_chan:
			return [
				{'$match': {'channel': self.channel}},
				{'$addFields': {'meta': self.conform_meta()}},
				{'$project': {'channel': 0, 'meta.activity.channel': 0}}
			]
		return []

	def t1(self) -> list[dict[str, Any]]:

		return [
			{'$match': {'channel': self.channel}},
			{'$addFields': {'meta': self.conform_meta()}},
			{'$project': {'channel': 0, 'meta.activity.channel': 0}}
		]


	def t2(self) -> list[dict[str, Any]]:

		return [
			{'$match': {'channel': self.channel}},
			{'$addFields': {'meta': self.conform_meta()}},
			{'$project': {'channel': 0, 'meta.activity.channel': 0}}
		]


	def t3(self) -> list[dict[str, Any]]:

		return [
			{'$match': {'channel': self.channel}},
			{'$addFields': {'meta': self.conform_meta()}},
			{'$project': {'channel': 0, 'meta.activity.channel': 0}}
		]


	def filter_journal(self, arg: str) -> dict[str, Any]:

		return {
			'$filter': {
				'input': f"${arg}",
				'as': "out",
				'cond': {
					'$or': [
						{"$not": ["$$out.channel"]},
						{'$eq': ["$$out.channel", self.channel]},
						{
							'$and': [
								{'$isArray': "$$out.channel"},
								{'$in': [self.channel, "$$out.channel"]}
							]
						}
					]
				}
			}
		}


	def get_meta_cases(self, arg: str) -> list[dict[str, Any]]:

		return [
			{'case': {'$eq': [self.channel, f"{arg}.channel"]}, 'then': arg},
			{
				'case': {
					'$and': [
						{'$isArray': f"{arg}.channel"},
						{'$in': [self.channel, f"{arg}.channel"]}
					]
				},
				'then': arg
			}
		]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/view/FrozenValuesDict.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                18.12.2019
# Last Modified Date:  18.12.2019
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.view.ReadOnlyDict import ReadOnlyDict

def ro(arg):
	if arg.__class__ is list:
		return (*arg,)
	if arg.__class__ is FrozenValuesDict:
		return ReadOnlyDict(arg)
	return arg

class FrozenValuesDict(dict):
	"""
	Dict that recursively casts all values to immutable structures.
	Note that the dict itself is not Immutable
	"""

	def __setitem__(self, key, value):
		if isinstance(value, list):
			# cast to tuple
			super().__setitem__(key, (*[ro(el) for el in value],))
		elif isinstance(value, dict):
			super().__setitem__(key, ReadOnlyDict(value))
		else:
			super().__setitem__(key, value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/NoViewGenerator.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.08.2022
# Last Modified Date:  17.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Iterable
from collections.abc import Generator
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.t3.stage.BaseViewGenerator import BaseViewGenerator, T, T3Send


class NoViewGenerator(BaseViewGenerator[T]):

	def __init__(self,
		unit: AbsT3ReviewUnit,
		buffers: Iterable[AmpelBuffer],
		stock_updr: MongoStockUpdater
	) -> None:

		super().__init__(unit_name = unit.__class__.__name__, stock_updr = stock_updr)
		self.buffers = buffers
		self.View = unit._View

	def __iter__(self) -> Generator[T, T3Send, None]:
		for ab in self.buffers:
			yield ab # type: ignore
